class WrongAnnotationPlacement(Exception):
    """
    It throws if the annotation is set over a function that takes arguments, or over a class/class method. When using
    the @test annotation with the data provider, it will be thrown if the function does not accept one argument.
    """
    pass


class DuplicateNameException(Exception):
    """
    It will be thrown if the provider name is already in the set, you need to rename the data provider method or give
    it a name through the parameter.
    """
    pass


class UnknownProviderName(Exception):
    """
    It will be thrown if, after the formation of the test run (test-suite), the provider with the specified name is not
    found.
    """
    pass


class TestBrokenException(Exception):
    """
    It will be thrown to notify of problems with the test, not related to assert.
    """
    pass


class TestIgnoredException(Exception):
    """
    It will be thrown to notify that the test is ignored (fixtures failed or only_if).
    """
    pass


class ExceptionWrapper(Exception):
    """
    The wrapper for exceptions to which you can then pass another exception. Used with waiting_exception.
    """

    def __init__(self):
        self.value = None
        self.message = 'Expect exception, but none raised!'
        self.args = (self.message,)
        self.type = type(self)

    def __str__(self):
        return str(self.value) if self.value else self.message

    def set_value(self, exception: Exception):
        self.value = exception
        self.args = exception.args
        self.message = self.args[0]
        self.type = type(exception)



from typing import Dict, List, Tuple, Callable, Iterable

from .timer import Timer
from .basic_test import Test
from .basic_group import TestGroup


class TestSuite:
    """
    The run class, test suite, always in a single instance (singleton), includes all sets.
    If there is at least 1 test, then it includes 1 group.
    """
    instance = None

    # The dictionary pairs name of group - TestGroup
    groups: Dict[str, TestGroup] = {}

    # Data providers (available to all tests in all sets)
    providers: Dict[str, Callable[[None], Iterable]] = {}

    # Cache for providers, store name-tuple of values
    cache: Dict[str, Tuple] = {}

    # Container for cached providers, stores only names!
    cached: List[str] = []

    # Lists of functions which execute before and after run
    before: List[Callable] = []
    after: List[Callable] = []
    name: str = 'Default Test Suite'

    # Flags that preliminary functions fell
    is_before_failed: bool = False

    # The flag that the final functions should be performed regardless of the result of preliminary
    always_run_after: bool = False

    # Timer for suite duration
    timer = Timer()

    def __new__(cls):
        if not cls.instance:
            cls.instance = super(TestSuite, cls).__new__(cls)
        return cls.instance

    @classmethod
    def add_before(cls, func: Callable):
        cls.before.append(func)

    @classmethod
    def add_after(cls, func: Callable):
        cls.after.append(func)

    @classmethod
    def get_instance(cls):
        return TestSuite()

    @classmethod
    def get_or_create(cls, group_name: str) -> TestGroup:
        """
        It creates and returns a test group by name.
        :param group_name: is the name of group
        :return: TestGroup
        """
        if group_name not in cls.groups:
            cls.groups[group_name] = TestGroup(group_name)
        return cls.groups[group_name]

    @classmethod
    def filter_groups(cls, groups: List[str]):
        """
        Leaves in the test_suite only those groups, whose names match the ones listed
        :param groups: list of group names
        :return: None
        """
        cls.groups = {name: tests for name, tests in cls.groups.items() if name in groups}

    @classmethod
    def filter_tests(cls, filter_by_name: str):
        """
        Leaves in the test_suite only those test whose names contains parameter (no full match necessary)
        :param filter_by_name: string, test stays in suite if its name contains this parameter
        :return: None
        """
        for group in cls.groups.values():
            group.tests = [test for test in group.tests if filter_by_name in test.name]

    @classmethod
    def is_empty(cls) -> bool:
        """
        It returns whether the suite is empty (no tests in any of the groups).
        :return: True if there are no tests
        """
        return all([group.is_empty() for group in cls.groups.values()])

    @classmethod
    def tests_count(cls) -> int:
        """
        Returns the total number of tests. Only tests with the parameter enabled=True get into the set.
        :return: is the number of the tests
        """
        if cls.is_empty():
            return 0
        return sum([group.tests_count() for group in cls.groups.values()])

    @classmethod
    def success(cls) -> List[Test]:
        """
        Returns the list of successful tests.
        :return:
        """
        return cls._test_result_of('success')

    @classmethod
    def failed(cls) -> List[Test]:
        """
        Returns the list of fell tests.
        :return:
        """
        return cls._test_result_of('failed')

    @classmethod
    def broken(cls) -> List[Test]:
        """
        Returns the list of broken tests (fell by exception, not by assert).
        :return:
        """
        return cls._test_result_of('broken')

    @classmethod
    def ignored(cls) -> List[Test]:
        """
        Returns the list of ignored tests (unsuccessful preliminary functions).
        :return:
        """
        return cls._test_result_of('ignored')

    @classmethod
    def _test_result_of(cls, name: str) -> List[Test]:
        return [test for group in cls.groups.values() for test in group.tests_by_status(name)]

    @classmethod
    def start_suite(cls):
        cls.timer.start()

    @classmethod
    def stop_suite(cls):
        cls.timer.stop()

    @classmethod
    def suite_duration(cls) -> float:
        return cls.timer.duration

    @classmethod
    def info(cls) -> Dict:
        """
        Statistics of whole test-suite
        :return: dict of main parameters
        """
        return {'name': cls.name, 'groups': tuple(cls.groups.keys()), 'providers': tuple(cls.providers.keys()),
                'tests_count': cls.tests_count(), 'duration': cls.suite_duration()}



import traceback

from .basic import Listener
from ..basic_test import Test
from ..basic_suite import TestSuite
from checking.helpers.others import short
from checking.helpers.others import format_seconds
from checking.helpers.others import print_splitter_line
from checking.helpers.exception_traceback import get_trace_filtered_by_filename


class DefaultListener(Listener):
    """
    Listener by default for standard behaviour during for run events.
    The user can write his own listeners by the class blueprint.
    """

    def __init__(self, verbose: int):
        super().__init__(verbose)
        self.counts = 0

    def on_before_suite_failed(self, test_suite):
        super().on_before_suite_failed(test_suite)
        print(f'Before suite "{test_suite.name}" failed! Process stopped!')

    def on_suite_starts(self, test_suite: TestSuite):
        super().on_suite_starts(test_suite)
        print(f'Starting suite "{test_suite.name}"')

    def on_empty_suite(self, test_suite: TestSuite):
        super().on_empty_suite(test_suite)
        print('No tests were found! Stopped...')

    def on_dry_run(self, test_suite: TestSuite):
        super().on_dry_run(test_suite)
        print_splitter_line()
        print("DRY RUN MODE! No real tests will be executed! All fixtures will be ignored!")
        print_splitter_line()

    def on_suite_stop_with_max_fail(self, max_fail: int):
        super().on_suite_stop_with_max_fail(max_fail)
        print_splitter_line()
        print(f"ATTENTION! Suite stops because of reaching maximum count of failed tests={max_fail}")
        print_splitter_line()

    def on_suite_ends(self, test_suite: TestSuite):
        super().on_suite_ends(test_suite)
        if not self.verbose:
            print()
        print()
        print("=" * 30)
        elapsed = format_seconds(test_suite.suite_duration())
        success_count = len(test_suite.success())
        f_count = len(test_suite.failed())
        b_count = len(test_suite.broken())
        i_count = len(test_suite.ignored())
        all_count = f_count + b_count + i_count + success_count
        print(f'Total tests: {all_count}, success tests: {success_count}, failed tests: {f_count}, broken tests: '
              f'{b_count}, ignored tests: {i_count}')
        print(f'Time elapsed: {elapsed}.')
        if self.verbose == 3:
            if f_count:
                print(f'\nFailed tests are:')
                for failed_test in test_suite.failed():
                    print(' ' * 4, failed_test, Listener._get_test_arg_short_without_new_line(failed_test))
            if b_count:
                print(f'\nBroken tests are:')
                for broken_test in test_suite.broken():
                    print(' ' * 4, broken_test, Listener._get_test_arg_short_without_new_line(broken_test))
            if i_count:
                print(f'\nIgnored tests are:')
                for ignored_test in test_suite.ignored():
                    print(' ' * 4, ignored_test)
        print("=" * 30)

    def on_success(self, test: Test):
        super().on_success(test)
        if not self.verbose:
            print('.', end='')
            self.counts += 1
            if self.counts > 100:
                print('')
                self.counts = 0
        elif self.verbose > 1:
            print_splitter_line()
            add_ = Listener._get_test_arg_short_without_new_line(test)
            print(f'Test "{test}" {add_} SUCCESS!')

    def on_broken(self, test: Test, exception_: Exception):
        super().on_broken(test, exception_)
        self._failed_or_broken(test, exception_, 'broken')

    def on_failed(self, test: Test, exception_: Exception):
        super().on_failed(test, exception_)
        self._failed_or_broken(test, exception_, 'failed')

    def on_ignored(self, test: Test, fixture_type: str):
        super().on_ignored(test, fixture_type)
        add_ = '' if not test.argument else f'[{short(test.argument)}]'
        print(f'Because of fixture "{fixture_type}" tests "{test}" {add_} was IGNORED!')
        print_splitter_line()

    def on_fixture_failed(self, group_name: str, fixture_type: str, exception_: Exception):
        super().on_fixture_failed(group_name, fixture_type, exception_)
        print_splitter_line()
        print(f'Fixture {fixture_type} "{group_name}" failed!')
        for tb in (e for e in traceback.extract_tb(exception_.__traceback__)):
            print(f'File "{tb.filename}", line {tb.lineno}, in {tb.name}')
        print(exception_)

    def on_ignored_with_provider(self, test: Test):
        super().on_ignored_with_provider(test)
        print(f'Provider "{test.provider}" for {test} not returns iterable or empty!'
              f' All tests with provider were IGNORED!')

    def on_test_starts(self, test: Test):
        super().on_test_starts(test)

    def on_ignored_by_condition(self, test: Test, exc: Exception):
        super().on_ignored_by_condition(test, exc)
        add_ = '' if not test.argument else f'[{short(test.argument)}]'
        if type(exc) is SystemExit:
            print(f'Test "{test}" {add_} ignored because of sys.exit() call inside function!')
        else:
            print(f'Test "{test}" {add_} ignored because of condition ({test.only_if.__module__}.{test.only_if})!')

    def _failed_or_broken(self, test: Test, exception_: Exception, _result: str):
        _letter = f'{_result.upper()}!'
        if not self.verbose:
            print(_letter[0], end='')
            self.counts += 1
            if self.counts > 100:
                print('')
                self.counts = 0
        else:
            if self.verbose > 0:
                print_splitter_line()
            add_ = Listener._get_test_arg_short_without_new_line(test)
            print(f'Test "{test}" {add_} {_letter}')
            print(get_trace_filtered_by_filename(exception_))
            print(exception_)



from typing import Any, Union, Sequence, Sized

from .helpers.others import short
from .exceptions import TestBrokenException


def is_true(obj: Any, message: str = None):
    """
    Checking of the object for truth.
    :param obj: is the object for checking
    :param message: optional message, recommended to use it
    :return: None
    """
    if not obj:
        _message = _mess(message)
        raise AssertionError(f'{_message}Expected True, but got False! ')


def is_false(obj: Any, message: str = None):
    """
    Checking an object for non-truth (false).
    :param obj: is the object for checking
    :param message: optional message, recommended to use it
    :return: None
    """
    if obj:
        _message = _mess(message)
        raise AssertionError(f'{_message}Expected False, but got True')


def equals(expected: Any, actual: Any, message: str = None):
    """
    It compares the equality of two objects.
    :param expected: is the expected object
    :param actual: is the actual object
    :param message: is the message to be indicated when verification fails
    :return: None
    :raises AssertionError if the objects are not equal, indicating the objects and their types
    """
    if (expected is actual) or expected == actual:
        return
    _message = _mess(message)
    raise AssertionError(f'{_message}Objects are not equal!\n'
                         f'Expected:"{short(expected, 150)}" <{type(expected).__name__}>\n'
                         f'Actual  :"{short(actual, 150)}" <{type(actual).__name__}>!')


def not_equals(expected: Any, actual: Any, message: str = None):
    """
    It checks that objects are not equal.
    :param expected: is the expected object
    :param actual: is the actual object
    :param message: is the message to be indicated when verification fails
    :return: None
    :raises AssertionError if objects are equal
    """
    if (expected is actual) or expected == actual:
        _message = _mess(message)
        raise AssertionError(f'Objects are equal: ({short(expected)}, {short(actual)})!')


def is_none(obj: Any, message: str = None):
    """
    It checks that the object is None, the inverse function for checking not_none.
    :param obj: is the checked object
    :param message: is the message to be indicated when verification fails
    :return: None
    :raises AssertionError with type of the object
    """
    _message = _mess(message)
    if obj is not None:
        raise AssertionError(f'{_message}Object {short(obj)}<{type(obj).__name__}> is not None!')


def is_not_none(obj: Any, message: str = None):
    """
    It checks that the object is not None, the inverse function for is_none.
    :param obj: is the checked object
    :param message: is the message to be indicated when verification fails
    :return: None
    :raises AssertionError
    """
    _message = _mess(message)
    if obj is None:
        raise AssertionError(f'{_message}Unexpected None!')


def test_fail(message: str = None):
    """
    Forced failure of the test, can be used in rare conditions instead of checking obviously wrong conditions.
    :param message: is the optional message
    :return: None
    """
    raise AssertionError(message if message else 'Test was forcibly failed!')


def test_brake(message: str = None):
    """
    Forcibly bring the test to a broken state, can be used in rare conditions instead of throwing exceptions.
    :param message: is the optional message
    :return: None
    """
    raise TestBrokenException(message if message else 'Test was forcibly broken!')


def contains(part: Any, whole: Any, message: str = None):
    """
    Checks that one object is part of (enters) another. Similar to check a in b.
    :param part: is the object-part that is part of the whole
    :param whole: is the entity that contains a part
    :param message: is the optional message
    :return: None
    :raises AssertionError if one object is part of the second
    :raises TestBrokenException if whole is not iterable or objects cannot be checked for content, for instance
    1 in '123'
    """
    __contains_or_not(part, whole, message=message)


def not_contains(part: Any, whole: Any, message: str = None):
    """
    Checks that one object is not part (not included) of the second. Similar to checking a not in b.
    :param part: is the part object that is part of the whole
    :param whole: is the integer object that contains part
    :param message: is the optional message
    :return: None
    :raises AssertionError if one object is part of the second
    :raises TestBrokenException if whole is not iterable or objects cannot be checked for content, for instance
    1 not in '123'
    """
    __contains_or_not(part, whole, is_contains=False, message=message)


def is_zero(actual: Union[int, float]):
    """
    Function checks that the object is 0.
    :param actual: int or float argument
    :return: None
    """
    _check_argument_is_number(actual, 'is_zero')
    if actual != 0:
        raise AssertionError(f'"{short(actual)}"<{type(actual).__name__}> is not equal to zero!')


def is_positive(actual: Union[int, float, Sequence]):
    """
    Function checks that the object is bigger than 0, if it is Sequence(list, str, etc.) checks that length is positive.
    :param actual: int, float or Sequence argument
    :return: None
    """
    if type(actual) in (int, float):
        if actual <= 0:
            raise AssertionError(f'{actual} is not positive!')
    else:
        if _get_length_if_sized(actual) <= 0:
            raise AssertionError(f'Length of "{short(actual)}"<{type(actual).__name__}> is not positive!')


def is_negative(actual: Union[int, float]):
    """
    Function checks that the object is smaller than 0.
    :param actual: int or float argument
    :return: None
    """
    _check_argument_is_number(actual, 'is_negative')
    if actual >= 0:
        raise AssertionError(f'"{short(actual)}"<{type(actual).__name__}> is not negative!')


def is_empty(container: Sized, message: str = None):
    """
    Function checks the container is empty (len==0)
    :param message: optional message
    :param container: any object which has size
    :return: None
    """
    _message = _mess(message)
    if _get_length_if_sized(container):
        raise AssertionError(f'{_message}"{short(container)}"<{type(container).__name__}> is not empty!')


def is_not_empty(container: Sized, message: str = None):
    """
    Function checks the container is NOT empty (len>0)
    :param message: optional message
    :param container: any object which has size
    :return: None
    """
    _message = _mess(message)
    if not _get_length_if_sized(container):
        raise AssertionError(f'{_message}"{short(container)}"<{type(container).__name__}> is empty!')


def __contains_or_not(part, whole, is_contains: bool = True, message: str = None):
    try:
        if is_contains and part in whole:
            return
        if not is_contains and part not in whole:
            return
    except TypeError as e:
        if 'requires' in e.args[0]:
            raise TestBrokenException(f'Object "{short(part)}" <{type(part).__name__}> and '
                                      f'"{short(whole)}"<{type(whole).__name__}> are of different types and '
                                      f'cant be check for contains!')
        raise TestBrokenException(
            f'"{short(whole)}"<{type(whole).__name__}> is not iterable and cant be check for contains!')
    _message = _mess(message)
    add_ = 'is a' if not is_contains else 'is not'
    raise AssertionError(f'{_message}Object "{short(part)}" <{type(part).__name__}>, {add_} part of \n'
                         f'"{short(whole, 150)}"<{type(whole).__name__}>!')


def _mess(message: str) -> str:
    return f'{message}\n' if message else ''


def _get_length_if_sized(container: Sized) -> int:
    """
    Returns length of container or raise Exception
    :param container: Sized object
    :return: length of object
    :raises: TestBrokenException if object has ho length
    """
    try:
        return len(container)
    except TypeError:
        raise TestBrokenException(f'"{short(container)}"<{type(container).__name__}> '
                                  f'has no len and cant be checked for emptiness!')


def _check_argument_is_number(actual, name: str):
    if type(actual) not in (int, float):
        raise TestBrokenException(f"Only int or float types allowed for {name}," +
                                  f" but got '{short(actual)}'<{type(actual).__name__}>")



from typing import Any, List, Tuple

from .calls import Call
from .interfaces import Observer
from .wrapper import AttributeWrapper


class Spy(Observer):
    """
    The test-double (spy), which replaces the desired object. His attributes return None, and the methods do not do
    anything if unless otherwise indicated, but all of the calls are fixed. This class in used due to make sure in
    the call of respectively functions with arguments.
    """

    def __init__(self, obj: Any = None):
        super().__init__()
        self.chain: List[Call] = []
        self._returns = None
        self._raises = None
        if obj is not None:
            for name in dir(obj):
                if name == '__dict__':
                    continue
                if callable(getattr(obj, name)):
                    if name != '__class__':
                        setattr(self, name, AttributeWrapper(name, self))
                    else:
                        setattr(self, name, self.__class__)
                else:
                    setattr(self, name, None)
        self.basic = obj

    def notify(self, _call: Call):
        self.chain.append(_call)

    def __call__(self, *args, **kwargs):
        if self._raises is not None:
            raise self._raises
        self.chain.append(Call('', *args, **kwargs))
        return self._returns

    def __str__(self):
        if self.basic is None:
            return f'Empty Test Spy'
        return f'Test spy of the "{self.basic}" {type(self.basic)}'

    def all_calls(self) -> List[Call]:
        """
        Returns list of Call objects (all method calls of the spied object)
        :return: List[Call]
        """
        return self.chain

    def was_called(self) -> bool:
        """
        Returns True if spy object was called itself
        :return: bool
        """
        return self.was_function_called('')

    def was_called_with_argument(self, arg: Any) -> bool:
        """
        Returns True if spy object was called itself with exact argument
        :param arg: Any argument to look for
        :return: bool
        """
        return self.was_function_with_argument_called('', arg)

    def returns(self, result: Any):
        """
        If spy object will be called itself return result
        :param result: any type to return when call
        :return: None
        """
        self._returns = result

    def raises(self, exception_object: Exception):
        """
        Raise an exception if object will be called itself (not its methods!)
        :param exception_object: exception object to raise
        :return: None
        """
        self._raises = exception_object

    def was_function_called(self, name: str) -> bool:
        """
        Returns True if exact function/method was called on spied object
        :param name: name of the function/method
        :return: bool
        """
        return any([e for e in self.chain if e.name == name])

    def was_function_with_argument_called(self, name: str, arg: Any) -> bool:
        """
        Returns True if exact function/method was called with exact argument on spied object
        :param name: name of the function/method
        :param arg: any argument
        :return: bool
        """
        if not self.was_function_called(name):
            return False
        return any([e for e in self.chain if e.name == name and arg in e.args])

    def was_exact_function_called(self, name, *args, **kwargs):
        call = Call(name, *args, **kwargs)
        return any([e for e in self.chain if e == call])

    def all_calls_args(self) -> List[Tuple]:
        """
        Returns all called function/method arguments
        :return: List[Tuple]
        """
        return [e.args for e in self.chain]

    def all_calls_args_flatten(self) -> List[Any]:
        """
        Returns flat list of all arguments of all called functions
        :return: List[Any]
        """
        return [arg for call in self.chain for arg in call.args]


class TestDouble(Spy):
    """
    The full test-double (twin of the object), the main difference with Spy is behaviour. Behaviour stays the same
    as original object has, but all calls fixed and you can change return result of the methods.
    This class in used due to make sure in the call of respectively functions with arguments.
    """

    def __init__(self, obj: Any = None):
        super().__init__()
        if obj is not None:
            for name in dir(obj):
                attr = getattr(obj, name)
                if callable(attr):
                    if name == '__dict__':
                        continue
                    if name != '__class__':
                        wrapper = AttributeWrapper(name, self)
                        wrapper.use_function(attr)
                        setattr(self, name, wrapper)
                    else:
                        setattr(self, name, self.__class__)
                else:
                    setattr(self, name, attr)
        wrapper = AttributeWrapper('len', self)
        setattr(self, 'len', wrapper)
        wrapper = AttributeWrapper('bool', self)
        setattr(self, 'bool', wrapper)
        wrapper = AttributeWrapper('iter', self)
        setattr(self, 'iter', wrapper)
        self.basic = obj

    def __str__(self):
        return f'Test Double of the "{self.basic}" {type(self.basic)}'

    def __len__(self):
        if self.len._return is not None or self.len._function:
            return self.len()
        self.len()
        return len(self.basic)

    def __bool__(self):
        if self.bool._return is not None or self.bool._function:
            return self.bool()
        self.bool()
        return bool(self.basic)

    def __iter__(self):
        if self.iter._return is not None or self.iter._function:
            return self.iter()
        self.iter()
        return iter(self.basic)



from typing import List, Callable


class TestCase:
    """
    Parent for test and group of tests (test set) has the common methods.
    """
    __slots__ = ('name', 'before', 'after', 'is_before_failed', 'always_run_after')

    def __init__(self, name: str):
        self.name = name
        # The list of the functions that execute before test/group
        self.before: List[Callable] = []
        # The list of the functions that execute after test/group
        self.after: List[Callable] = []
        # The flag that the preliminary operation has fallen, in this case the function will not start after the test
        self.is_before_failed: bool = False
        # The flag of obligatory start of function after the test, even if the preliminary ones has fell down
        self.always_run_after: bool = False

    def add_before(self, func: Callable):
        self.before.append(func)

    def add_after(self, func: Callable):
        self.after.append(func)


from typing import Callable, Any, Optional

from .timer import Timer
from .basic_case import TestCase
from ..exceptions import TestIgnoredException


class Test(TestCase):
    """
    The class which representing test, the main point of test run.
    """
    __slots__ = ('name', 'before', 'after', 'is_before_failed', 'always_run_after', 'provider', 'retries', 'priority',
                 'test', 'group', 'group_name', 'argument', 'timeout', 'only_if', 'description', 'timer', 'status',
                 'reason', 'report_params')

    def __init__(self, name: str, test: Callable):
        """
        Object initialization requires not only name, but a function reference, which is a test.
        :param name: is the name of the test
        :param test: is the function that will be launch at the test executes
        """
        super().__init__(name)
        # Function to run (the test itself)
        self.test = test
        self.group_name: str = '__main__'
        self.group = None
        # Argument for data-provider run
        self.argument: Any = None
        # Time in seconds to finish test
        self.timeout: int = 0
        # Function-predicate, if return False - test will not runs
        self.only_if: Optional[Callable] = None
        # Description of the test
        self.description = test.__doc__
        # The name of the provider for future delivery of data to the test
        self.provider = None
        # The number of the test run attempts
        self.retries: int = 1
        # Test priority where 0 is highest
        self.priority: int = 0
        # Timer for start, end and duration parameters of the test
        self.timer = Timer()
        # Status of the test
        self.status: str = 'created'
        # Raised exception
        self.reason: Optional[Exception] = None
        # Parameters to attach to html report (not used anythere else)
        self.report_params = {}

    def set_group(self, group):
        self.group = group
        self.group_name = group.name

    def run(self):
        """
        The launch of the test (functions mark tests annotation).
        :return: None
        """
        if self.only_if:
            if not self.only_if():
                raise TestIgnoredException()
        self.timer.start()
        if self.provider is not None:
            self.test(self.argument)
        else:
            self.test()

    def stop(self, exception_: Optional[BaseException] = None):
        """
        Stop the test - just get the time and save status and exception. After that put test to results of the group.
        :param exception_: If None, then test is succeed.
        :return: None
        """
        self._stop()
        self.reason = exception_
        if not exception_:
            self.status = 'success'
        elif type(exception_) in (TestIgnoredException, SystemExit):
            self.status = 'ignored'
            self.timer.start_time = -1
            self.timer._end_time = -1
            self.timer.duration = 0
        elif type(exception_) is AssertionError:
            self.status = 'failed'
        else:
            self.status = 'broken'
        self._put_to_group_results()

    def _stop(self):
        """
        Save end time and calculate duration of the test
        :return: None
        """
        self.timer.stop()

    def _put_to_group_results(self):
        if self.group:
            self.group.add_result(self)

    def __str__(self):
        description = self.description if self.description else ''
        if description:
            description = description.replace('\n', '').replace('  ', '')
            description = f" ('{description}')"
        return f'{self.group_name}.{self.name}{description}'

    def clone(self):
        """
        Cloning the instance
        :return: a new Test
        """
        clone = Test(self.name, self.test)
        for attr in self.__slots__:
            if attr != 'report_params':
                setattr(clone, attr, getattr(self, attr))
        return clone

    def duration(self) -> float:
        return self.timer.duration

    def info(self) -> dict:
        """
        Statistics of the test
        :return: dict of main parameters
        """
        return {'group': self.group_name, 'name': self.name, 'description': self.description, 'priority': self.priority,
                'argument': self.argument, 'timeout': self.timeout, 'retries': self.retries, 'status': self.status,
                'reason': self.reason, 'duration': self.duration()}



import logging
from threading import currentThread

from .basic import Listener
from ..basic_test import Test
from ..basic_suite import TestSuite
from checking.helpers.others import short, format_seconds, str_date_time


class DefaultFileListener(Listener):
    """
    Listener by default for logging all events to local file.
    The user can write his own listeners by the class blueprint.
    """

    def __init__(self, log_level: str = logging.DEBUG, name: str = None, use_time: bool = True):
        """
        Init all parameters for logger
        :param log_level: level of logging, DEBUG by default
        :param name: name of the result file, by default is 'test_suite'
        :param use_time: if True, appends date and time to file name of the log.
        Example: test_suite_2020-04-09_17-51-29.log
        """
        super().__init__(0)
        name_ = 'test_suite' if not name else name
        str_time = str_date_time()
        time_ = f'_{str_time}' if use_time else ''
        format_ = '%(asctime)-15s Thread[%(threadName)s] %(levelname)s: %(message)s'
        handler = logging.FileHandler(f'{name_}{time_}.log', 'w', encoding='utf-8')
        logging.basicConfig(level=log_level, format=format_, handlers=(handler,))

    def on_suite_starts(self, test_suite: TestSuite):
        super().on_suite_starts(test_suite)
        logger = logging.getLogger()
        logger.info(f'Test-suite "{test_suite.name}" started!')

    def on_dry_run(self, test_suite: TestSuite):
        super().on_dry_run(test_suite)
        logger = logging.getLogger()
        logger.warning("DRY RUN MODE! No real tests will be executed! All fixtures will be ignored!")

    def on_suite_ends(self, test_suite: TestSuite):
        super().on_suite_ends(test_suite)
        elapsed = format_seconds(test_suite.suite_duration())
        success_count = len(test_suite.success())
        f_count = len(test_suite.failed())
        b_count = len(test_suite.broken())
        i_count = len(test_suite.ignored())
        all_count = f_count + b_count + i_count + success_count
        logger = logging.getLogger()
        logger.info(f'Test-suite "{test_suite.name}" finished!')
        logger.info(f'Total tests: {all_count}, success tests: {success_count}, failed tests: {f_count}, broken tests: '
                    f'{b_count}, ignored tests: {i_count}')
        logger.info(f'Time elapsed: {elapsed}.')

    def on_test_starts(self, test: Test):
        super().on_test_starts(test)
        logger = logging.getLogger(currentThread().getName())
        logger.info(f'Test  "{test}" started!')

    def on_empty_suite(self, test_suite: TestSuite):
        super().on_empty_suite(test_suite)
        logger = logging.getLogger()
        logger.critical(f'Test-suite "{test_suite.name}" is empty! No tests found')

    def on_fixture_failed(self, group_name: str, fixture_type: str, exception_: Exception):
        super().on_fixture_failed(group_name, fixture_type, exception_)
        logger = logging.getLogger(currentThread().getName())
        logger.exception(f'Fixture {fixture_type} "{group_name}" failed!', exc_info=exception_)

    def on_ignored_with_provider(self, test: Test):
        super().on_ignored_with_provider(test)
        logger = logging.getLogger(currentThread().getName())
        logger.warning(f'Provider "{test.provider}" for {test} not returns iterable or empty! '
                       f'All tests with provider were IGNORED!')

    def on_success(self, test: Test):
        super().on_success(test)
        add_ = Listener._get_test_arg_short_without_new_line(test)
        logger = logging.getLogger(currentThread().getName())
        logger.info(f'Test "{test}" {add_} SUCCESS!')

    def on_failed(self, test: Test, exception_: Exception):
        super().on_failed(test, exception_)
        add_ = Listener._get_test_arg_short_without_new_line(test)
        logger = logging.getLogger(currentThread().getName())
        logger.exception(f'Test "{test}" {add_} FAILED!', exc_info=exception_)

    def on_broken(self, test: Test, exception_: Exception):
        super().on_broken(test, exception_)
        add_ = Listener._get_test_arg_short_without_new_line(test)
        logger = logging.getLogger(currentThread().getName())
        logger.exception(f'Test "{test}" {add_} BROKEN!', exc_info=exception_)

    def on_ignored(self, test: Test, fixture_type: str):
        super().on_ignored(test, fixture_type)
        add_ = '' if not test.argument else f'[{short(test.argument)}]'
        logger = logging.getLogger(currentThread().getName())
        logger.warning(f'Because of fixture "{fixture_type}" tests "{test}" {add_} was IGNORED!')

    def on_ignored_by_condition(self, test: Test, exc: Exception):
        super().on_ignored_by_condition(test, exc)
        add_ = '' if not test.argument else f'[{short(test.argument)}]'
        logger = logging.getLogger(currentThread().getName())
        if type(exc) is SystemExit:
            logger.warning(f'Test "{test}" {add_} ignored because of sys.exit() call inside function!')
        else:
            logger.warning(f'Test "{test}" {add_} ignored because of condition'
                           f' ({test.only_if.__module__}.{test.only_if})!')

    def on_before_suite_failed(self, test_suite):
        super().on_before_suite_failed(test_suite)
        logger = logging.getLogger()
        logger.critical(f'Before suite "{test_suite.name}" failed! Process stopped!')



from typing import List, Callable
from inspect import isfunction

from checking.asserts import *
from checking.helpers.exception_traceback import get_trace_filtered_by_filename


class SoftAssert:
    """
    The class for applying soft checks, that is, checks that do not stop the test immediately, but allow to perform
    other checks. This can be convenient for obtaining information about various parts of the system under test, for
    instance, for validating some of fields. It allows you to get information about all the fallen checks, and not about
    the first, as is the case with a simple test.
    An example:

    soft_assert = SoftAssert()
    soft_assert.check(lambda : equals(1, my_json['field1'], 'message'))
    soft_assert.check(lambda : equals('text', my_json['field2'], 'message'))
    soft_assert.contains(1,my_json['list'])
    soft_assert.assert_all()
    """

    def __init__(self, check_immediately: bool = False):
        """
        The initializer that accepts by default that there is no need to do checks immediately. It could be critical
        for tests, where the system state can change due to checks or other actions.
        :param check_immediately: is the flag for immediately checking the condition, if it is True, then the check will
        done right away, but the test does not fall, all results anyway will be displayed after assert_all() calling.
        """
        self.__funcs = []
        self.__result = []
        self.__check_immediately = check_immediately

    def check(self, lambda_):
        """
        Soft check that takes lambda as an argument executes later, when checking all conditions. If flag
        check_immediately=True, then checked (executed) immediately, but exceptions will be drop during the final run.
        It is recommended to use this method for obtaining the error with the trace.
        An example:

        soft_assert = SoftAssert()
        soft_assert.put(lambda : equals(1,2,'message'))

        :param lambda_: the lambda does not take any parameters
        :return: None
        """
        if not isfunction(lambda_):
            raise TestBrokenException(f'Argument for SoftAssert.check function must be function, preferably lambda!')
        if self.__check_immediately:
            self._check(lambda_)
        else:
            self.__funcs.append(lambda_)

    def assert_all(self):
        """
        Checking all conditions, which have been received before and display all list fallen checks with string
        indicating.
        Supposed, that it is the final action in any test.
        :return: None
        :raises AssertionError in the case of checks fall
        :raises Exception if there were other exceptions during checks execution
        """
        if not self.__check_immediately:
            for func in self.__funcs:
                self._check(func)
        if not self.__result:
            return
        raise AssertionError(self._create_message(self.__result))

    def equals(self, expected: Any, actual: Any, message: str = None):
        self.check(lambda: equals(expected, actual, message))

    def is_none(self, obj: Any, message: str = None):
        self.check(lambda: is_none(obj, message))

    def not_none(self, obj: Any, message: str = None):
        self.check(lambda: is_not_none(obj, message))

    def contains(self, part: Any, whole: Any, message: str = None):
        self.check(lambda: contains(part, whole, message))

    def not_contains(self, part: Any, whole: Any, message: str = None):
        self.check(lambda: not_contains(part, whole, message))

    def is_true(self, obj: Any, message: str = None):
        self.check(lambda: is_true(obj, message))

    def is_false(self, obj: Any, message: str = None):
        self.check(lambda: is_false(obj, message))

    @staticmethod
    def _create_message(exceptions: List[Exception]) -> str:
        message = '=' * 20
        message += '\nFAILED ASSERTS:\n'
        for exception in exceptions:
            message += "\n".join([get_trace_filtered_by_filename(exception), str(exception), f'{("-" * 20)}\n'])
        return message

    def _check(self, func: Callable):
        try:
            func()
        except Exception as ex:
            self.__result.append(ex)



from os import sep
from os import chdir
from os import mkdir
from os import path
from sys import _getframe
from datetime import datetime
from typing import List, Union
from checking.helpers.others import is_file_exists

from checking.classes.basic_test import Test
from checking.classes.basic_suite import TestSuite
from checking.helpers.exception_traceback import get_trace_filtered_by_filename as filtered

# Name for folder with results
FOLDER = 'test_results'


def add_text(name: str, value: str):
    """
    Add some text to test in report
    :param name: name of the parameter
    :param value: value of the parameter
    :return: None
    """
    _add_to_test(name, value)


def add_img(name: str, bytes_: bytes):
    """
    Add some picture(screenshot) to test in report
    :param name: name of parameter
    :param bytes_: bytes of picture (will be saved as png)
    :return: None
    """
    _add_to_test(name, bytes_)


def _add_to_test(name: str, value: Union[bytes, str]):
    """
    Looks for 5 frames deep in call stack to find current test and put data there
    :param name: name of parameter to add
    :param value: value of str or bytes
    :return: None
    """
    try:
        for _ in range(5):
            frame = _getframe(_)
            current_test = frame.f_locals.get('self')
            if current_test:
                if type(value) is bytes:
                    current_test.report_params[value] = str(name)
                else:
                    current_test.report_params[str(name)] = str(value)
                break
    except ValueError:
        pass  # ignored, just deletes frame
    finally:
        del frame


def generate(test_suite: TestSuite):
    """
    The only public method to generate html report with all test results
    :param test_suite: suite to save results to report
    :return: None
    """
    path_to_template = is_file_exists.__globals__['__spec__'].origin
    full_path = path.join(path.split(path_to_template)[0], 'index_real.html')
    css_path = path.join(path.split(path_to_template)[0], 'materialize.css')
    roboto_path = path.join(path.split(path_to_template)[0], 'roboto.css')
    if not is_file_exists(FOLDER):
        mkdir(FOLDER)
        chdir(FOLDER)
        if not is_file_exists('css'):
            mkdir('css')
        chdir('..')
    with open(f'{FOLDER}{sep}index.html', 'wt') as file, open(full_path) as template:
        file.write(template.read())
    with open(f'{FOLDER}{sep}css{sep}materialize.css', 'wt') as file, open(css_path) as template:
        file.write(template.read())
    with open(f'{FOLDER}{sep}css{sep}roboto.css', 'wt') as file, open(roboto_path) as template:
        file.write(template.read())
    html = _generate_html(test_suite)
    _write_file(html)


def _create_info(html: str, suite: TestSuite):
    """
    Generates Info section (header) of the report.
    :param html: sting of the html report
    :param suite: suite to get results from
    :return: None
    """
    per_col = '#2e7d32'
    percent = len(suite.success()) / (suite.tests_count() / 100) if suite.tests_count() else 0.0
    if percent < 99:
        per_col = '#ff6f00'
    if percent < 75:
        per_col = '#d50000'
    start = datetime.fromtimestamp(suite.timer.start_time).strftime('%Y-%m-%d %H:%M:%S')
    end = datetime.fromtimestamp(suite.timer.end_time).strftime('%Y-%m-%d %H:%M:%S')
    html = html.replace('#suite_name', suite.name). \
        replace('#total_groups', str(len(suite.groups))). \
        replace('#total_tests', str(suite.tests_count())). \
        replace('#success_tests', str(len(suite.success()))). \
        replace('#failed_tests', str(len(suite.failed()))). \
        replace('#broken_tests', str(len(suite.broken()))). \
        replace('#ignored_tests', str(len(suite.ignored()))). \
        replace('#percent', f"<b style='color: {per_col}'>0.0 %</b>"). \
        replace('#total_time', f"{suite.suite_duration():.2} seconds ({start} - {end})")
    return html


def _generate_html(test_suite: TestSuite) -> List[str]:
    """
    Generates html in list of strings
    :param test_suite: suite to get all results from
    :return: list of strings (html code)
    """
    with open(f'{FOLDER}{sep}index.html') as template:
        base = ''.join(template.readlines())
    base = _create_info(base, test_suite)
    html_lines = [base, ]
    if test_suite.is_empty():
        html_lines.append("<div id='empty'>Suite is empty! There are no tests!</div>\n")
        html_lines.append('</body>\n</html>')
        return html_lines
    count = 1
    for group in test_suite.groups:
        group_time = float(sum(t.duration() for t in test_suite.groups.get(group).test_results))
        results = test_suite.groups.get(group).test_results
        succ = len(test_suite.groups.get(group).tests_by_status('success'))
        html_lines.append(
            f"<h4 id='id_g_{count}' style='cursor: pointer;'>Group '{group}' (elapsed {group_time:.2} seconds), "
            f"success tests {succ}/{len(results)}:\n"
            f"\t<script>document.querySelector('#id_g_{count}')."
            f"addEventListener('click', opclose_sibling('#id_g_{count}'))</script>\n</h4>\n"
            f"<ol style='display: none;'>\n")
        for test in results:
            _add_test_info(test, html_lines, count)
            count += 1
        html_lines.append('</ol>\n')
    _add_all_listeners(html_lines, count)
    html_lines.append('\n</body>\n</html>')
    return html_lines


def _write_file(lines: list):
    with open(f'{FOLDER}{sep}index.html', 'wt') as file:
        file.write(''.join(lines))


def _add_all_listeners(lines: List[str], count: int):
    a_l = [f"document.querySelector('#id_{_}').addEventListener('click',opclose('#id_{_}'));" for _ in range(1, count)]
    joined = '\n'.join(a_l)
    lines.append(f"<script>{joined}</script>")


def _add_test_info(test: Test, lines: List[str], count: int):
    """
    Add info about specific test
    :param test: test to ger info from
    :param lines: list of strings og html file
    :param count: id for html tag
    :return: None
    """
    time_ = 0.0 if test.duration() < 0.01 else test.duration()
    add_ = f'[{test.argument}]' if test.provider else ''
    traceback = '' if test.reason is None else ''.join([f"<p>{_}</p>" for _ in filtered(test.reason).split('\n')])
    st_col = '#2e7d32'
    if test.status == 'broken':
        st_col = '#ff6f00'
    if test.status == 'failed':
        st_col = '#d50000'
    if test.status == 'ignored':
        st_col = 'grey'
    rep_params = _get_rep_params(test, count)
    lines.append(
        f"\t<li id='id_{count}' style='cursor: pointer;'>Test '{test.name}' {add_}: elapsed time {time_:.2} seconds, "
        f"status <b style='color:{st_col}'>{test.status}</b>\n"
        f"\t\t<div style='display: none;'>\n<table style='width: 80%; margin-left: auto; margin-right: auto; font-family: 'Roboto', sans-serif;'>\n<tbody>\n"
        f"\t\t\t<tr><td><b>Description:</b></td><td>'{test.description}'</td></tr>\n"
        f"\t\t\t<tr><td><b>Argument:</b></td><td> {test.argument}</td></tr>\n"
        f"\t\t\t<tr><td><b>Attempt number:</b></td><td> {test.retries}</td></tr>\n"
        f"\t\t\t<tr><td><b>Status:</b></td><td><b style='color:{st_col}'>{test.status.upper()}</b></td></tr>\n"
        f"\t\t\t<tr><td><b>Duration:</b></td><td> {time_:.3} seconds</td></tr>\n</tbody>\n</table>\n"
        f"{traceback}\n"
        f"\t\t\t<p style='color: {st_col}'><b>Exception:</b> {str(test.reason).replace('<', '&lt;')}</p><hr/>\n"
        f"{rep_params}"
        f"\t\t</div>\n\t</li>\n")


def _get_rep_params(test, count):
    """
    Add info attached to test (text ot image)
    :param test: test to get info from
    :param count: id for element
    :return: None
    """
    rep_params = ''
    if not test.report_params:
        return ''
    for key, value in test.report_params.items():
        if type(key) is str:
            rep_params += f"\t\t\t<p>Report parameter <b>'{key}'</b>: {value}</p>\n"
        else:
            with open(f'{FOLDER}{sep}test_id_count_{count}.png', 'wb') as f:
                f.write(key)
            rep_params += f"\t\t\t<p>Report parameter <b>'{value}'</b>:</p>\n\t\t\t" \
                          f"<img src='test_id_count_{count}.png' alt='{value}'>\n"
    return rep_params



import os
import sys
import json
import argparse
import importlib
from typing import Dict, Union, List

from checking.runner import start
from checking.helpers.others import str_date_time
from checking.helpers.others import is_file_exists
from checking.classes.listeners.default import DefaultListener
from checking.classes.listeners.file_logger import DefaultFileListener

HOME_FOLDER = sys.path[0]
LOOK_FOR = ('import checking', 'from checking')
DEFAULT_LISTENERS = {'DefaultListener': DefaultListener, 'DefaultFileListener': DefaultFileListener}


def read_parameters_from_file(file_name_: str) -> Dict:
    """
    Reads the parameters from settings file *.json, returns as a parameter dictionary.
    """
    with open(file_name_, encoding='utf-8') as file:
        result = ''.join([line.rstrip() for line in file.readlines()])
    parameters = _get_default_params()
    parameters.update(json.loads(result))
    return parameters


def check_parameters(parameters: Dict):
    """
    Checks the validation of start parameters.
    :param parameters: the dictionary with parameters
    :return: None
    :raises ValueError if the parameters are invalid
    """
    schema = {bool: ['dry_run', 'random_order', 'generate_report'], str: ['suite_name', 'listener', 'filter_by_name'],
              int: ['threads', 'verbose', 'max_fail'],
              list: ['modules', 'groups'], dict: ['params']}
    for key, value in schema.items():
        for element in value:
            if element not in parameters:
                continue
            if type(parameters[element]) is not key:
                raise ValueError(f'{element.capitalize()} parameter must be {key}!')
    for name in schema.get(list):
        if name not in parameters:
            continue
        if any(filter(lambda x: type(x) is not str, parameters[name])):
            raise ValueError(f'{name.capitalize()} parameter must be list of strings (List[str])!')
    listener_ = parameters.get('listener')
    if listener_:
        if '.' not in listener_ and listener_ not in DEFAULT_LISTENERS:
            raise ValueError(f'Listener parameter must contains module and class name, like "my_module.MyListener" or'
                             f' it must be the name of one of the default listeners!')


def _get_default_params():
    parameters = {'suite_name': 'Default Test Suite', 'verbose': 0, 'groups': [], 'params': {}, 'listener': '',
                  'modules': [], 'threads': 1, 'dry_run': False, 'filter_by_name': '', 'random_order': False,
                  'max_fail': 0, 'generate_report': False}
    return parameters


def start_with_parameters(parameters: Dict):
    params_ = _get_default_params()
    params_.update(parameters)
    check_parameters(params_)
    modules = params_.get('modules')
    listener_ = params_.get('listener')
    try:
        if listener_ and modules:
            # if used custom listener  -add it module to filter
            if listener_ not in DEFAULT_LISTENERS:
                modules.append('.'.join(listener_.split('.')[:-1]))
        _walk_throw_and_import(modules)
        if listener_:
            # if listener is one of defaults - use it
            if listener_ in DEFAULT_LISTENERS:
                params_['listener'] = DEFAULT_LISTENERS[listener_](params_['verbose'])
            else:
                # or try to import it from filtered module
                class_ = _get_class_from_imported_modules(listener_)
                # Instantiate listener-class object
                params_['listener'] = class_(params_['verbose'])
    except Exception:
        print(f'Something wrong with importing!', file=sys.stderr)
        raise
    start(**params_)


def _get_class_from_imported_modules(listener_name: str):
    """
    Get class of test-listener from imported modules in system
    :param listener_name: full name like module.Listener
    :return: class for instantiation
    """
    pack = listener_name.split('.')[-2]
    cl_ = listener_name.split('.')[-1]
    keys = [key for key in sys.modules.keys() if key.endswith(pack)]
    if not keys:
        raise ValueError(f'Cant find listener {listener_name}')
    mod = sys.modules.get(keys[0])
    return getattr(mod, cl_)


def _is_import_in_file(file_name_: str) -> bool:
    """
    If there checking import within the file.
    :param file_name_: is the name of the module
    :return: True, if asserts imports inside the module
    """
    with open(file_name_, encoding='utf-8') as file:
        for line in file:
            line = line.rstrip().replace('  ', ' ')
            if any([element in line for element in LOOK_FOR]):
                return True
    return False


def _walk_throw_and_import(filter_modules: List[str] = None):
    """
    The functions runs throw all of the modules in current and nested folder, looks up and imports modules, where is a
    mention checking, thereby forming the test-suite.
    :return: None
    """
    for root, dirs, files in os.walk(HOME_FOLDER, topdown=True):
        if root.endswith(os.sep + 'checking') or (os.sep + 'checking' + os.sep in root):
            continue
        files = (file for file in files if file.endswith(".py"))
        for file in (f for f in files if _is_import_in_file(f'{root}{os.sep}{f}')):
            file_name_ = file.replace('.py', '')
            if filter_modules:
                if not _is_in_filter_list(filter_modules, file_name_, root):
                    continue
            try:
                package_ = root.replace(HOME_FOLDER, '').replace(os.sep, '')
                name = package_ + '.' + file_name_ if package_ else file_name_
                importlib.import_module(name, package=HOME_FOLDER)
            except ModuleNotFoundError:
                sys.path.append(root)
                importlib.import_module(file_name_)


def _is_in_filter_list(filter_modules: List[str], file_name_: str, root: str) -> bool:
    filtered = [mod for mod in filter_modules if mod.endswith(f'{file_name_}')]
    if not filtered:
        return False
    if '.' in filtered[0]:
        filtered_with_package = [mod for mod in filtered
                                 if root.replace(os.sep, '.').endswith('.'.join(mod.split('.')[:-1]))]
        if not filtered_with_package:
            return False
    return True


def _generate_options():
    """
    Generates default options<date+time>.json file at current folder
    :return: None
    """
    with open(f'options_{str_date_time()}.json', 'wt') as file:
        json.dump(_get_default_params(), file)


def _get_file_name(arg: Union[str, None]) -> str:
    name = 'options.json'
    if arg:
        name = arg
        if not name.endswith('.json'):
            raise ValueError('Only <name>.json files allowed! And it must contains valid json, '
                             'you can generate it with -g option')
    return name


def _get_arg_dict(arg: Union[str, None]) -> Dict:
    dic_ = {}
    if arg:
        if '=' in arg:
            key, value = arg.split("=")
            dic_[key] = value
        else:
            dic_[arg] = None
    return dic_


def parse_arguments() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument('-o', '--options_file', help="File with the options to run. "
                                                     "If specified, -d and -f options will be ignored!")
    parser.add_argument('-f', '--filter_test', help="Filter tests by name, only tests, whose names"
                                                    " contain this value will run!")
    parser.add_argument('-a', '--arg', help="Any argument for your test-suite")
    parser.add_argument('-r', '--random', help="Runs tests inside group in random order", action='store_const',
                        const=True)
    parser.add_argument('-d', '--dry_run', help="Dry-run for test-suite(run without actual functions executed)!",
                        action='store_const', const=True)
    parser.add_argument('-g', '--generate_options',
                        help="Not doing any work, just generates options.json in current folder!", action='store_const',
                        const=True)
    parser.add_argument('-m', '--max_fail', help="Maximum count of failed test, if reached then suite will stops")
    parser.add_argument('-R', '--generate_report',
                        help="Creates folder with html report in current folder!", action='store_const',
                        const=True)
    return parser.parse_args()


def _main_run(file_name_: str, p_: Dict, dry_run_: bool, filter_by_name_: str, random_order: bool, max_fail: int,
              generate_report: bool):
    # if options file exists get all from there
    if is_file_exists(file_name_):
        print(f"{file_name_} found! Work with it...")
        params_ = read_parameters_from_file(file_name_)
        if p_:
            params_.get("params").update(p_)
        start_with_parameters(params_)
    # or walk recursive and find all modules with tests
    else:
        if file_name_ == 'options.json':
            print(f"No options found! Starts to look for tests in all sub-folders")
            _walk_throw_and_import()
            start(verbose=3, threads=1, params=p_, dry_run=dry_run_, filter_by_name=filter_by_name_,
                  random_order=random_order, max_fail=max_fail, generate_report=generate_report)
        # if options file was specified and not exists, than stop
        else:
            raise ValueError(f"{file_name_} not found! Stopped")


def run():
    args = parse_arguments()
    if args.generate_options:
        _generate_options()
    else:
        file_name = _get_file_name(args.options_file)
        params = _get_arg_dict(args.arg)
        dry_run = args.dry_run if args.dry_run else False
        random_order = args.random if args.random else False
        filter_by_name = args.filter_test if args.filter_test else ''
        max_fail = int(args.max_fail) if args.max_fail else 0
        generate_report = args.generate_report if args.generate_report else False
        _main_run(file_name, params, dry_run, filter_by_name, random_order, max_fail, generate_report)


if __name__ == '__main__':
    run()



from typing import Iterable, Sized, Type

from checking.asserts import *
from checking.helpers.others import short


class FluentAssert:
    """
    The class for flexible and readable checks, which can be assembled in chains (sharing if desired AND).
    If there is failure of one of the checks, the test falls and other conditions are not checked.
    An example:

    verify(my_list).is_not_none().AND.is_a(list).AND.contains(2).AND.is_sorted()

    """

    def __init__(self, obj: Any):
        """
        During initialization receive an object, which need to check. It is not intended to be a class or type!
        :param obj:
        """
        self.__actual = obj
        self._t = type(self.__actual)
        # Строковое представление объекта в сокращенном виде (не более 50 символов)
        self.str = short(self.__actual)
        # Ссылка на самого себя для читаемой последовательности проверок
        self.AND = self

    @property
    def size(self):
        """
        Switch all following checks to the length of actual object
        :return: FluentAssert
        :raises: TestBrokenException if object has no length
        """
        self._check_has_len(self.__actual)
        return FluentAssert(len(self.__actual))

    def has_attribute(self, name: str):
        """
        Check, if actual object has attribute with given name
        :param name: string name of the attribute
        :return: FluentAssert
        """
        if not hasattr(self.__actual, name):
            raise AssertionError(f'"{self.str}"{self._t} has no attribute "{short(name)}"')
        return self

    def attribute(self, name: str):
        """
        Switch all following checks to attribute of the object or raise exception, if object has no attribute with
        given name
        :param name: name of the attribute
        :return: FluentAssert
        :raises: AssertionError if object has no attribute
        """
        self.has_attribute(name)
        return FluentAssert(getattr(self.__actual, name))

    def same_as(self, obj: Any):
        """
        Check, if the instance and the current one are the same. Similarly of checking a is b.
        :param obj: is the instance for checking
        :return: None
        """
        if self.__actual is not obj:
            raise AssertionError(f'"{self.str}"{self._t} is not "{short(obj)}"{type(obj)}')
        return self

    def not_same_as(self, obj: Any):
        """
        Check, if the instance and the current one are not the same. Similarly of checking a is not b.
        :param obj: is the instance for checking
        :return: None
        """
        if self.__actual is obj:
            raise AssertionError(f'"{self.str}"{self._t} is the same as "{short(obj)}"{type(obj)}')
        return self

    def child_of(self, type_: Type):
        """
        Check, if the instance is a concrete class(type) or his child.
        :param type_: is the type for checking
        :return: None
        """
        self._check_is_type(type_)
        if issubclass(type(self.__actual), type_):
            return self
        raise AssertionError(f'"{self.str}"{self._t} is not sub-class of {type_}')

    def type_is(self, type_: Type):
        """
        Check, if the  actual object is instance of given class.
        :param type_: class to check
        :return: FluentAssert
        :raises: AssertionError if object is not instanse of given type
        """
        self._check_is_type(type_)
        if isinstance(self.__actual, type_):
            return self
        raise AssertionError(f'"{self.str}"{self._t} is not of type {type_}')

    def is_none(self):
        is_none(self.__actual)
        return self

    def is_not_none(self):
        is_not_none(self.__actual)
        return self

    def is_true(self):
        is_true(self.__actual)
        return self

    def is_false(self):
        is_false(self.__actual)
        return self

    def equal(self, obj: Any):
        equals(self.__actual, obj)
        return self

    def not_equal(self, obj: Any):
        not_equals(self.__actual, obj)
        return self

    def is_empty(self):
        is_empty(self.__actual)
        return self

    def is_not_empty(self):
        is_not_empty(self.__actual)
        return self

    def is_positive(self):
        is_positive(self.__actual)
        return self

    def is_negative(self):
        is_negative(self.__actual)
        return self

    def is_zero(self):
        is_zero(self.__actual)
        return self

    def less_than(self, obj: Any):
        """
        Checking that the checking object is less than current.
        :param obj: is the object for comparing
        :return:
        """
        self._check_same_type(obj)
        if self.__actual >= obj:
            raise AssertionError(f'"{self.str}" is not less than "{short(obj)}"!')
        return self

    def less_or_equal_to(self, obj: Any):
        self._check_same_type(obj)
        if self.__actual > obj:
            raise AssertionError(f'"{self.str}" is not less and not equal to "{short(obj)}"!')
        return self

    def greater_than(self, obj: Any):
        self._check_same_type(obj)
        if self.__actual <= obj:
            raise AssertionError(f'"{self.str}" is not greater than "{short(obj)}"!')
        return self

    def greater_or_equal_to(self, obj: Any):
        self._check_same_type(obj)
        if self.__actual < obj:
            raise AssertionError(f'"{self.str}" is not equal and not greater than "{short(obj)}"!')
        return self

    def equal_to_length_of(self, obj: Sized):
        self._check_has_len(obj)
        self.equal(len(obj))
        return self

    def less_than_length_of(self, obj: Sized):
        self._check_has_len(obj)
        self.less_than(len(obj))
        return self

    def greater_than_length_of(self, obj: Sized):
        self._check_has_len(obj)
        self.greater_than(len(obj))
        return self

    def is_sorted(self, reverse_order: bool = False):
        """
        Checking, that object is sorted (it can be used just to list/tuple/str and another Sequence).
        :param reverse_order: if it is True then check sort in reverse (from the biggest to the lowest)
        :return:
        """
        if not isinstance(self.__actual, Sequence):
            raise TestBrokenException(f'Only sequences can be checked for sorted, not {self._t}')
        smaller = lambda a, b: a <= b
        bigger = lambda a, b: a >= b
        check = smaller if not reverse_order else bigger
        for index, element in enumerate(self.__actual):
            if index == len(self.__actual) - 1:
                break
            if not check(element, self.__actual[index + 1]):
                raise AssertionError(f'Object "{self.str}" is not sorted!')
        return self

    def contains(self, obj: Any):
        contains(obj, self.__actual)
        return self

    def not_contains(self, obj: Any):
        not_contains(obj, self.__actual)
        return self

    def contains_in_any_order(self, obj: Iterable):
        if not isinstance(obj, Iterable):
            raise TestBrokenException(f'Only Iterables can be argument here, not {type(obj)}')
        for element in obj:
            contains(element, self.__actual)
        return self

    def is_in(self, container: Any):
        contains(self.__actual, container)
        return self

    def is_not_in(self, container: Any):
        not_contains(self.__actual, container)
        return self

    def _check_same_type(self, second):
        if self._t is not type(second):
            raise TestBrokenException('To compare "less" or "greater" both arguments must be the same type!')

    def _check_is_type(self, obj):
        if type(obj) is not type:
            raise TestBrokenException(f'Argument "{obj}"{type(obj)} is not type!')

    def _check_has_len(self, obj):
        try:
            len(obj)
        except TypeError:
            raise TestBrokenException(f'There is no length for "{short(obj)}"{type(obj)}')


def verify(obj: Any) -> FluentAssert:
    return FluentAssert(obj)



from .stub import Stub
from .calls import Call
from .doubles import Spy
from .doubles import TestDouble
from .interfaces import Observer
from .wrapper import AttributeWrapper

__all__ = ('Spy', 'Call', 'Stub', 'TestDouble', 'Observer', 'AttributeWrapper')



from typing import Callable


class DataFile:
    """
    Wrapper for open text file, can change line on reading with function. Used for data-provider.
    """

    def __init__(self, file_name: str, encoding: str = 'UTF-8', map_function: Callable = None):
        self.file = open(file_name, encoding=encoding)
        self.map_ = map_function if map_function else lambda line: line

    def __iter__(self):
        return (self.map_(line) for line in self.file)

    def close(self):
        self.file.close()



from typing import Any, List, Callable

from .calls import Call
from .interfaces import Observer


class AttributeWrapper:
    """
    Wrapper for all methods of mocked object. Use it to change behaviour, returns something or raise exception
    on method calls.
    """

    def __init__(self, name: str, observer: Observer):
        self.name = name
        self._function = None
        self._return = None
        self._raise = None
        self._raise_condition = None
        self._calls = []
        self.observer = observer

    def __str__(self):
        return f'Wrapper for method "{self.name}" on object: {self.observer}"'

    def returns(self, arg: Any):
        """
        Specify what to return on object method call
        :param arg: any object
        :return: None
        """
        self._return = arg
        self._function = None

    def raises(self, exception: Exception, message: str = None):
        """
        Raises this exception on method call
        :param exception: Exception or its children
        :param message: Message for exception
        :return: None
        """
        self._raise = (exception, message)
        self._raise_condition = None

    def raises_if_args(self, predicate: Callable[[Any], bool], exception: Exception, message: str = None):
        """
        Raises this exception on method call, if predicate returns True on call arguments. If method will be call
        without arguments - no checks will be performed, and no exception raise.
        :param predicate: function to check call arguments
        :param exception: Exception to raise
        :param message: additional message
        :return: None
        """
        self._raise = (exception, message)
        self._raise_condition = predicate

    def __call__(self, *args, **kwargs):
        self._calls.append((args, kwargs))
        self.observer.notify(Call(self.name, *args, **kwargs))
        if self._raise:
            exception, message = self._raise
            if self._raise_condition and (args or kwargs):
                if self._raise_condition(*args, **kwargs):
                    raise exception(message)
            elif not self._raise_condition:
                raise exception(message)
        if self._function:
            return self._function(*args, **kwargs)
        return self._return

    def use_function(self, function: Callable):
        """
        Will call specified function on method call
        :param function: function to call
        :return: None
        """
        self._function = function

    def was_called(self) -> bool:
        return len(self._calls) > 0

    def call_count(self) -> int:
        return len(self._calls)

    def all_calls(self) -> List:
        return self._calls



import builtins
from io import StringIO, BytesIO
from inspect import ismodule
from contextlib import contextmanager
from typing import Any, Type, Union, List, Tuple, Set

from .exceptions import ExceptionWrapper
from .exceptions import TestBrokenException
from .classes.mocking.doubles import Spy
from .classes.mocking.write_wrapper import WriteWrapper


@contextmanager
def mock_builtins(function_name: str, func):
    """
    Mock of built-in functions like print, input and so on. After exiting the context manager, the original function
    regains its previous behavior.
    :param function_name: is the the name of one of the python built-in function
    :param func: is the replacement function, which will be called instead of the original
    :return:
    """
    if function_name == 'open':
        print(f'WARNING! For using mock with open function, please use mock_open. '
              f'Read the documentation at https://bitbucket.org/kotolex/atest/src')
    import builtins as b
    if not hasattr(b, function_name):
        raise TestBrokenException(f'No build-in function "{function_name}"!')
    temp_ = None
    try:
        temp_ = getattr(b, function_name)
        setattr(b, function_name, func)
        yield
    finally:
        if temp_:
            setattr(b, function_name, temp_)


@contextmanager
def mock_open(on_read_text: str = '', on_read_bytes: bytes = b'', new_line: str = '\n', raises: Exception = None):
    """
    Context manager for mocking open text ot byte file. Use it instead of mock_builtins('open', func)
    :param on_read_text: string to get when open text file in read mode
    :param on_read_bytes: bytes to get when open bytes file in read mode
    :param new_line: how to determine new line, '\n' by default, will be ignored for bytes
    :param raises: if need to raise exception on open file. If not None - other options will be ignored
    :return: list with values, which be pushed into it by write on open file
    :raises: ValueError if on_read_* not str or bytes
    """
    container = []

    def fake_open(*args, **kwargs):
        if raises:
            raise raises
        if (len(args) > 1 and 'w' in args[1]) or (kwargs.get("mode") and 'w' in kwargs.get("mode")):
            return WriteWrapper(container)
        if (len(args) > 1 and 'b' in args[1]) or (kwargs.get("mode") and 'b' in kwargs.get("mode")):
            return BytesIO(on_read_bytes)
        return StringIO(on_read_text, new_line)

    # If we raise on open, then no need to check arguments
    if raises is None:
        if type(on_read_text) is not str:
            raise ValueError("Parameter on_read_text must be str!")
        if type(on_read_bytes) is not bytes:
            raise ValueError('Parameter on_read_bytes must be bytes')
    temp_ = None
    try:
        temp_ = builtins.open
        builtins.open = fake_open
        yield container
    finally:
        if temp_:
            builtins.open = temp_


@contextmanager
def mock(module_: Any, function_name: str, func: Any):
    """
    Context manager for mocking (spoofing) any function or module attribute.
    :param module_: is the module object (not a name! it must be imported in the test)
    :param function_name: is the function name
    :param func: is the replacement function, but there may be an attribute
    :return:
    """
    if not ismodule(module_):
        raise TestBrokenException(f'"{module_} is not a module!')
    if not hasattr(module_, function_name):
        raise TestBrokenException(f'No function "{function_name} at module {module_}"!')
    temp_ = None
    try:
        temp_ = getattr(module_, function_name)
        setattr(module_, function_name, func)
        yield
    finally:
        if temp_:
            setattr(module_, function_name, temp_)


@contextmanager
def waiting_exception(exception: Type[Exception]):
    """
    The context manager to check if an exception is thrown during certain actions. An example:

    with waiting_exception(ZeroDivisionError) as exc:
        some_action()
    print(exc.message) # Displays a message from the exception

    :param exception: is the expected exception type, you cannot use BaseException, it is not recommended to use
    Exception (better use specific exception)
    :return: ExceptionWrapper context, which is initially empty, and when an exception is thrown, gets it in the value
    parameter
    :raises TestBrokenException if BaseException is used or not inheritors of Exception
    :raises AssertionError if the wrong exception is thrown, which was expected
    :raises ExceptionWrapper if no exceptions are raised
    """
    fake = ExceptionWrapper()
    try:
        if exception is BaseException:
            raise TestBrokenException('You must use concrete exception, except of BaseException!')
        if not issubclass(type(exception), type(Exception)):
            raise TestBrokenException(f'Exception or its subclasses expected, but got '
                                      f'"{exception}"<{type(exception).__name__}>')
        yield fake
    except TestBrokenException as e:
        raise e
    except exception as e:
        fake.set_value(e)
        return
    except Exception as e:
        raise AssertionError(f'Expect {exception}, but raised {type(e).__name__} ("{e}")')
    else:
        raise fake


@contextmanager
def no_exception_expected():
    """
    The context manager for situations where exceptions are not expected to raise is more explicit than just writing a
    test without an assertion. An example:

    with no_exception_expected():
        some_action()

    :return: None
    :raises AssertionError if the exception (any descendant of Exception) nevertheless falls
    """
    try:
        yield
    except Exception as e:
        raise AssertionError(f'Expect no exception, but raised {type(e).__name__} ("{e}")!')



from .asserts import *
from .context import *
from .runner import start
from .annotations import *
from .classes.mocking import *
from .runner import common_parameters
from .classes.fluent_assert import verify
from .classes.soft_assert import SoftAssert
from .classes.listeners.file_logger import DefaultFileListener

__all__ = ['start', 'common_parameters', 'SoftAssert', 'Spy', 'TestDouble', 'DefaultFileListener', 'DATA_FILE', 'CONTAINER',
           'equals', 'is_none', 'is_not_none', 'waiting_exception', 'test_fail', 'test_brake', 'no_exception_expected',
           'contains', 'verify', 'not_contains', 'not_equals', 'is_false', 'is_true', 'mock_builtins', 'mock',
           'is_zero', 'is_positive', 'is_negative', 'mock_open', 'is_empty', 'is_not_empty', 'Stub',
           'test', 'before', 'after', 'before_group', 'after_group', 'before_suite', 'after_suite', 'provider']



import os
import traceback
from typing import List, Tuple

SEPARATOR = os.path.sep


def _is_need_to_hide(name: str) -> bool:
    parts = (f'checking{SEPARATOR}asserts.py',
             f'checking{SEPARATOR}classes{SEPARATOR}',
             f'checking{SEPARATOR}runner.py{SEPARATOR}',
             f'{SEPARATOR}contextlib.py',)
    """
    Check the strings of traceback for the contents of the internal modules of the project and some modules of standard 
    library, to as not to display error traces (which the user is not interested in).
    :param name: is the name
    :return: True if it is the name of one of the internal modules.
    """
    return any([part in name for part in parts])


def get_trace(exception_: Exception) -> List[Tuple]:
    """
    Returns all trace of error as list, which consists with tuples (error line location, an error).
    :param exception_:  is the exception that whose trace is needed.
    :return: the list of tuples.
    """
    result = []
    for tb in (e for e in traceback.extract_tb(exception_.__traceback__)):
        first = f'File "{tb.filename}", line {tb.lineno}, in {tb.name}'
        second = f'-->    {tb.line}'
        result.append((first, second))
    return result


def get_trace_filtered_by_filename(exception_: Exception) -> str:
    """
    Returns string representation of the trace error, which has been filtered by module name, actually remain only
    strings which are not owned by the main 'checking' classes.
    :param exception_: the exception, whose stacktrace is needed
    :return: the string, where each note is defined \n
    """
    return '\n'.join([f'{a}\n{b}' for a, b in get_trace(exception_) if not _is_need_to_hide(a)])



from .interfaces import Observer
from .wrapper import AttributeWrapper


class Stub(Observer):
    """
    Stub is a test-double which main purpose is to give some data to application under test in cases when it need data
    to work and test properly. It is not for checking or assertions, just for separate application from data source.
    """
    def __init__(self, **kwargs):
        """
        You can mention all attributes here in key-value pairs. If you need to specify method or function use
        stub.method.returns(None) instead
        :param kwargs: key-value pairs for fill attributes of object
        """
        super().__init__()
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __getattr__(self, item):
        if item not in self.__dict__:
            self.__dict__[item] = AttributeWrapper(item, self)
        return self.__dict__[item]

    def __str__(self):
        return f'Stub object {self.__dict__}'

    def notify(self, call_):
        pass



from ..basic_test import Test
from ..basic_case import TestCase
from ..basic_suite import TestSuite
from checking.helpers.others import short


class Listener:
    """
    The parent of listeners of tests. Actually, the user can write their own listener and change the behaviour by
    default, by the way, his listener should be an inheritor of this class.
    """

    def __init__(self, verbose: int = 0):
        self.verbose = verbose

    def on_suite_starts(self, test_suite: TestSuite):
        """
        It calls at the start of the run, after checking the providers.
        :param test_suite: is TestSuite
        :return: None
        """
        pass

    def on_suite_stop_with_max_fail(self, max_fail: int):
        """
        It calls when max_fail parameter was specified and reached, so suite stops
        :param max_fail: count of failed test
        :return:
        """
        pass

    def on_suite_ends(self, test_suite: TestSuite):
        """
        It calls at the end of the run, when all tests finished and fixtures.
        :param test_suite: is TestSuite
        :return: None
        """
        pass

    def on_test_starts(self, test: TestCase):
        """
        It calls before test starts (after preliminary fixtures, but before the test).
        :param test: is the instance of the test with all parameters
        :return: None
        """
        pass

    def on_empty_suite(self, test_suite: TestSuite):
        """
        It calls before the run stops because of there are no tests in any of the groups.
        :param test_suite: is TestSuite
        :return: None
        """
        test_suite.start_suite()
        test_suite.stop_suite()

    def on_dry_run(self, test_suite: TestSuite):
        """
        It calls if suite is run in dry mode (no real tests executes)
        :param test_suite: is TestSuite
        :return: None
        """
        pass

    def on_fixture_failed(self, group_name: str, fixture_type: str, exception_: Exception):
        """
        It calls when a fixture failed to start (before/after).
        :param group_name: is the name of the group of tests
        :param fixture_type: is the name of fixture
        :param exception_: fell exception
        :return: None
        """
        pass

    def on_ignored_with_provider(self, test: Test):
        """
        It calls when the function which marked @data does not return Iterable and all of tests, which tied to the
        provider is ignored.
        :param test: is TestCase
        :return: None
        """
        pass

    def on_success(self, test: Test):
        """
        It calls when the test finished successfully.
        :param test: is TestCase
        :return: None
        """
        pass

    def on_failed(self, test: Test, exception_: Exception):
        """
        It calls if an assert falls at test (embedded or from checking.asserts module)
        :param test: is TestCase
        :param exception_: is fell assert
        :return: None
        """
        pass

    def on_broken(self, test: Test, exception_: Exception):
        """
        It calls when the test is 'broken', actually, it falls with an exception, but not by assert.
        :param test: is TestCase
        :param exception_: is fell assert
        :return: None
        """
        pass

    def on_ignored(self, test: Test, fixture_type: str):
        """
        It calls when the test has been ignored due to fallen fixture (before).
        :param test: is TestCase
        :param fixture_type: is the name of the fixture, called ignore
        :return: None
        """
        pass

    def on_ignored_by_condition(self, test: Test, exc: Exception):
        """
        It calls when the test has been ignored by statement (only_if).
        :param test: is TestCase
        :param exc: is the exception
        :return: None
        """
        pass

    def on_before_suite_failed(self, test_suite):
        """
        It calls at the falling of @before_suite fixture.
        :param test_suite: is TestSuite
        :return: None
        """
        pass

    @staticmethod
    def _get_test_arg_short_without_new_line(test: Test):
        if test.provider:
            short_ = short(test.argument).replace('\n', '')
            return f'[{short_}]'
        return ''



import glob
from typing import Any
from datetime import datetime


def short(obj: Any, width: int = 45) -> str:
    """
    Reduces the string representation to a specific length, replacing all characters after the specified length with
    three dots ([...]) plus symbols count to end.
    If the string representation is shorter than the specified length, it returns it unchanged.
    :param obj: String or object[Any]
    :param width: Maximum string representation length
    :return: shortened string
    """
    str_ = str(obj)
    len_ = len(str_) - width
    if len_ > 0:
        return str_[:width] + f'[...+{len_}]'
    return str_


def print_splitter_line():
    """
    Displays a dash separator in the form of a 10 dash (for separating test results)
    :return: None
    """
    print('-' * 10)


def is_file_exists(file_name: str) -> bool:
    """
    Tells, if the file with name exists
    :return: True, if the file exists
    """
    return len(glob.glob(file_name)) == 1


def str_date_time():
    """
    Generates current date and time for using at file names
    :return: str like 2020-02-22_12_58_58
    """
    return datetime.strftime(datetime.now(), '%Y-%m-%d_%H-%M-%S')


def fake(*args):
    """
    Stub function that does nothing
    :param args: ignored params
    :return: None
    """
    pass


def format_seconds(seconds: float) -> str:
    """
    Creates string representation of seconds count, which format it to hours and/or minutes if possible
    :param seconds: amount of seconds
    :return: str
    """
    if not seconds // 60:
        return f'{seconds:.2f} seconds'
    hours = ''
    if seconds // 3600:
        hours = f'{int(seconds // 3600)} hour(s) '
        seconds = seconds - ((seconds // 3600) * 3600)
    minutes = int(seconds // 60)
    seconds = seconds - minutes * 60
    return f'{hours}{minutes} minute(s) and {seconds:.2f} seconds'


if __name__ == '__main__':
    print(format_seconds(4135))



class Call:
    """
    The class which represents a single function call, stores its name and call arguments
    """

    def __init__(self, name: str, *args, **kwargs):
        self.name = name
        self.args = args
        self.kwargs = kwargs

    def __str__(self):
        args = self.args if self.args else 'no args'
        kwargs = self.kwargs if self.kwargs else 'no keyword args'
        return f'Call of "{self.name}" with {args}, {kwargs}'

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        if other is None:
            return False
        if type(other) != type(self):
            return False
        return self.name == other.name and self.args == other.args and self.kwargs == other.kwargs



class WriteWrapper:
    """
    Wrapper for mock_open, push all writes to list
    """

    def __init__(self, container: list):
        self.c = container
        self.opened = True

    def flush(self):
        if not self.opened:
            raise IOError('Attempt to write on closed wrapper!')

    def write(self, n):
        if not self.opened:
            raise IOError('Attempt to write on closed wrapper!')
        self.c.append(n)

    def writelines(self, __lines):
        for line in __lines:
            self.write(line)

    def close(self):
        self.opened = False

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="checking",
    version="0.7.2",
    author="Lex Draven",
    author_email="lexman2@yandex.ru",
    description="A small library for testing your code, made by tester for testers",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://bitbucket.org/kotolex/atest/",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



from typing import Callable
from threading import Thread, Event

from checking.classes.basic_test import Test
from checking.exceptions import TestBrokenException


class ExceptionThread(Thread):
    """
    The sub-class of the thread, for catching exceptions, which do not catch by the usual approach with multi-threaded
    execution.
    Saves exceptions within Event container.
    """

    def __init__(self, func: Callable, event: Event):
        super().__init__(target=func, daemon=True)
        self.func = func
        self.daemon = True
        self.event = event

    def run(self):
        try:
            self.func()
        except Exception as e:
            self.event.attr = e


def run_with_timeout(test: Test):
    event = Event()
    t = Thread(target=_start_test_at_daemon_thread, args=(test, event), daemon=True)
    t.start()
    t.join(test.timeout)
    # If the thread is alive, then it had not over by time-out.
    if t.is_alive():
        # Put the notification that it is time to abort the work and give up the exception.
        event.set()
        raise TestBrokenException(f'Time ({test.timeout} seconds) is over for test "{test}"!')
    # If there is an attribute, then the exception inside the thread-demon, then raise it.
    if hasattr(event, 'attr'):
        raise event.attr


def _start_test_at_daemon_thread(test: Test, event: Event):
    t = ExceptionThread(test.run, event)
    t.start()
    # Not yet notified throw the event or the thread-demon has been working now, then work
    while not event.is_set():
        if not t.is_alive():
            break



from random import shuffle
from typing import List, Callable

from .basic_test import Test
from .basic_case import TestCase


class TestGroup(TestCase):
    """
    The test set or group of tests is all tests from one module, or all tests are contained from one group, but it is
    not a test-suite (the set of all sets). Test-suite might be only one and contains all sets of tests, a test-suite is
    used to creating during testing starts. If there is at least 1 test, then a test set will be created and placed
    in a test-suite, during start.
    """

    __slots__ = ('name', 'before', 'after', 'is_before_failed', 'always_run_after', 'tests', 'before_all', 'after_all',
                 'test_results')

    def __init__(self, name: str):
        super().__init__(name)
        # The list of tests, i.e instances of 'Test' class
        self.tests: List[Test] = []
        # The list of functions that executed before each test
        self.before_all: List[Callable] = []
        # The list of functions that executed after each test
        self.after_all: List[Callable] = []
        # The list of the results of the run of this set
        self.test_results: List[Test] = []

    def add_test(self, test: Test):
        """
        Adding the test in the set, wherein receives the name of the group, lists of preliminaries and subsequents
        functions.
        :param test: is the instance of 'Test' class
        :return: None
        """
        test.set_group(self)
        test.before = self.before_all
        test.after = self.after_all
        self.tests.append(test)

    def add_before_test(self, func: Callable):
        """
        Adding one more function in preliminary pull, for executing before each test.
        :param func: is the function
        :return: None
        """
        self.before_all.append(func)

    def add_after_test(self, func: Callable):
        """
        Adding one more function in final pull, for executing after each test.
        :param func: is the function
        :return: None
        """
        self.after_all.append(func)

    def add_result(self, test: Test):
        """
        Adding the test in appropriate section of the results dictionary.
        :param test: is the instance of the TestCase
        :param result: is the result, by default it is successful
        :return: None
        """
        self.test_results.append(test)

    def is_empty(self):
        """
        Checking, if the set is empty.
        :return: True, if there are no tests inside the set
        """
        return len(self.tests) == 0

    def tests_count(self):
        """
        Returns the number of tests. If tests have been executed and there are results, it returns the sum of tests from
        results, otherwise it returns the number of 'TestCase; within 'tests' list. It has done due to at the time of
        starting the tests it is not know how many of them will be in total, thanks to restarts (retries) and/or to
        providers.
        :return: the number of the tests
        """
        return len(self.test_results) if self.test_results else len(self.tests)

    def sort_test_by_priority(self):
        self.tests = sorted(self.tests, key=lambda t: t.priority)

    def shuffle_tests(self):
        """
        Change test order randomly, ignoring priority
        :return: None
        """
        shuffle(self.tests)

    def tests_by_status(self, status: str) -> List[Test]:
        return [test for test in self.test_results if test.status == status]



from typing import Callable, Any, Dict, List
from concurrent.futures import ThreadPoolExecutor

from .helpers.others import fake
from .classes.basic_test import Test
from .classes.basic_case import TestCase
from .classes.basic_suite import TestSuite
from .classes.basic_group import TestGroup
from checking.helpers.report import generate
from .classes.listeners.basic import Listener
from .classes.exc_thread import run_with_timeout
from .classes.listeners.default import DefaultListener
from .exceptions import UnknownProviderName, TestIgnoredException

# Tests listener
_listener: Listener
# flag for running all test, sets to False to stop suite (when failed count reached)
_can_run = True
# maximum of fails for test in suite, if 0 then works till the end of tests
_max_fail = 0
# actual failed tests count
_actual_failed_count = 0

# Common parameters for whole test-suite
common_parameters: Dict[str, Any] = {}


def start(verbose: int = 0, listener: Listener = None, groups: List[str] = None, params: Dict[str, Any] = None,
          threads: int = 1, suite_name: str = 'Default Test Suite', dry_run: bool = False, filter_by_name: str = None,
          random_order: bool = False, max_fail: int = 0, generate_report: bool = False, **kwargs):
    """
    The main function of tests start

    :param suite_name: name of the test-suite
    :param listener: is test listener, DefaultListener is used by default. If set, then the verbose parameter is ignored
    (the one in the listener is used).
    :param verbose: is the report detail, 0 - briefly (only dots and 1 letter), 1 - detail, indicating only failed
    tests, 2 - detail, indicating successful and fallen, 3 - detail and at the end, a list of fallen and broken ones
    If not between 0 and 3, then 0 is accepted
    :param groups: is the list of group names to run, to run only the tests you need
    :param params: is the dictionary of parameters available in all tests (general run parameters)
    :param threads: is the number of threads to run tests by default is 1. Each group can run in a separate thread if
    necessary. This is an experimental feature and it can be useful only for tests NOT performing any complex
    calculations (CPU bound). It is best to use this parameter (more than 1) for tests related to the use of I / O
    operations - disk work, network requests. Obey the GIL!
    :param dry_run: if True runs test-suite with fake function except of real tests and fixtures, can be useful to
    find out order, number of tests, params of provider etc.
    :param filter_by_name if specified - runs only tests with name containing this parameter
    :param random_order if specified - runs tests inside each group in random order
    :param max_fail if greater than 0, than suite will stops, when reach that count of failed tests
    :param generate_report if specified - creates html report with the results in test folder
    :return: None
    """
    verbose = 0 if verbose not in range(4) else verbose
    if type(max_fail) is int and max_fail > 0:
        global _max_fail
        _max_fail = max_fail
    # If a listener is specified, then use it, otherwise by default
    global _listener
    _listener = listener if listener else DefaultListener(verbose)
    test_suite = TestSuite.get_instance()
    test_suite.name = suite_name
    if groups:
        test_suite.filter_groups(groups)
    if filter_by_name:
        test_suite.filter_tests(filter_by_name)
    # If there are no tests, then stop
    if test_suite.is_empty():
        _listener.on_empty_suite(test_suite)
        _listener.on_suite_ends(test_suite)
        if generate_report:
            generate(test_suite)
        return
    if params:
        common_parameters.update(params)
    if threads < 1:
        threads = 1
    if dry_run:
        _dry_run(test_suite)
    # Check if all used provider names are found
    _check_data_providers(test_suite)
    _run(test_suite, threads, random_order, generate_report)


def _dry_run(test_suite):
    """
    Clear all fixtures and replace real test with fake function
    :param test_suite: is TestSuite
    :return: None
    """
    _listener.on_dry_run(test_suite)
    test_suite.before = []
    test_suite.after = []
    for group in test_suite.groups.values():
        group.before = []
        group.after = []
        for test in group.tests:
            test.before = []
            test.after = []
            test.test = fake


def _run(test_suite: TestSuite, threads: int = 1, random_order: bool = False, generate_report: bool = False):
    for group in test_suite.groups.values():
        if random_order:
            group.shuffle_tests()
        else:
            group.sort_test_by_priority()
    # Create a pool of threads to run tests, by default 1 thread execution
    pool = ThreadPoolExecutor(max_workers=threads)
    test_suite.start_suite()
    _listener.on_suite_starts(test_suite)
    try:
        # If the fixture before the test suite has fallen, then will not run the tests (fixture after the test suite
        # will be executed with the appropriate flag)
        if not _run_before_suite(test_suite):
            return
        for group in test_suite.groups.values():
            # If there are no tests in the group, then skip it
            if not group.tests:
                continue
            pool.submit(_run_group_before_and_after_at_separate_thread, group)
        pool.shutdown(wait=True)
        _run_after(test_suite)
    finally:
        test_suite.stop_suite()
        _listener.on_suite_ends(test_suite)
        if generate_report:
            generate(test_suite)


def _run_group_before_and_after_at_separate_thread(group: TestGroup):
    # If the fixture before the group fell, then we do not run the tests
    if not _run_before_group(group):
        return
    _run_all_tests_in_group(group)
    _run_after(group)


def _run_before_suite(test_suite: TestSuite) -> bool:
    """
    Run fixtures before whole suite.
    :param test_suite: TestSuite
    :return: True if fixture is not fail, False otherwise
    """
    _run_before(test_suite)
    if test_suite.is_before_failed:
        _listener.on_before_suite_failed(test_suite)
        if test_suite.always_run_after:
            _run_after(test_suite)
        return False
    return True


def _run_before_group(group: TestGroup) -> bool:
    """
    Runs before group fixtures
    :param group: TestGroup
    :return: True if fixtures not failed, False otherwise
    """
    _run_before(group)
    if group.is_before_failed:
        for test in group.tests:
            test.stop(TestIgnoredException('Before module/group failed!'))
            _listener.on_ignored(test, 'before module/group')
        if group.always_run_after:
            _run_after(group)
        return False
    return True


def _run_test_with_provider(test):
    """
    Run test which has provider
    :param test: Test
    :return: None
    """
    test_suite = TestSuite.get_instance()
    provider = test.provider
    # If provider values is in the cache - get it from there
    generator = _provider_next(provider) if provider not in test_suite.cache else test_suite.cache[provider]
    try:
        is_any_value_provides = False
        need_to_cache = provider in test_suite.cached and provider not in test_suite.cache
        list_of_arguments = []
        for param in generator:
            is_any_value_provides = True
            if not _can_run:
                break
            clone = test.clone()
            clone.argument = param
            if need_to_cache:
                list_of_arguments.append(param)
            is_one_of_before_test_failed = _run_test_with_before_and_after(clone, False)
            if is_one_of_before_test_failed:
                print(f'Because of "before_test" all tests for {test} with data provider '
                      f'"{test.provider}" was IGNORED!')
                break
        # If no values at provider - ignore test
        if not is_any_value_provides:
            test.stop(TestIgnoredException(f'No values at provider {test.provider}'))
            _listener.on_ignored_with_provider(test)
        else:
            if need_to_cache:
                test_suite.cache[provider] = tuple(list_of_arguments)
    except TypeError as e:
        if 'is not iterable' not in e.args[0]:
            print(e)
            raise
        else:
            test.stop(TestIgnoredException(f'Error with provider {test.provider}'))
            _listener.on_ignored_with_provider(test)


def _run_all_tests_in_group(group: TestGroup):
    """
    Run all test of the group, after sorting them by priority
    :param group: TestGroup
    :return: None
    """
    is_one_of_before_test_failed = False
    for test in group.tests:
        if not _can_run:
            break
        if test.provider:
            _run_test_with_provider(test)
        else:
            is_one_of_before_test_failed = _run_test_with_before_and_after(test, is_one_of_before_test_failed)


def _run_test_with_before_and_after(test: Test, is_before_failed: bool) -> bool:
    if not is_before_failed:
        _run_before(test)
    else:
        test.is_before_failed = True
    if test.is_before_failed:
        test.stop(TestIgnoredException('Before test failed!'))
        _listener.on_ignored(test, 'before test')
        return True
    for retry in range(test.retries):
        clone = test.clone()
        if retry > 0:
            clone.name = clone.name + f' ({retry})'
        result = _run_test(clone)
        if result:
            break
    _run_after(test)
    return False


def _provider_next(provider_name: str) -> Any:
    """
    Gets provider functions return or yield(generator) and iterate it. If a close attribute exists (for files) try to
    close recourse after getting all elements.
    :param provider_name: name of the provider
    :return: generator
    """
    iter_ = TestSuite.get_instance().providers[provider_name]()
    for param in iter_:
        yield param
    # If it were file at provider - try to close it
    if hasattr(iter_, 'close'):
        try:
            iter_.close()
        except Exception as ex:
            print(ex)
            pass
            # explicitly ignore


def _run_test(test: Test) -> bool:
    """
    Run actual test
    :param test: Test (function to run)
    :return: True if test succeed, False otherwise
    """
    _listener.on_test_starts(test)
    try:
        if test.timeout:
            run_with_timeout(test)
        else:
            test.run()
        test.stop()
        _listener.on_success(test)
        return True
    except AssertionError as e:
        test.stop(e)
        _listener.on_failed(test, e)
        global _actual_failed_count
        if _max_fail and _actual_failed_count < _max_fail:
            _actual_failed_count += 1
        if _max_fail and _actual_failed_count >= _max_fail:
            global _can_run
            _can_run = False
            _listener.on_suite_stop_with_max_fail(_max_fail)
    except (TestIgnoredException, SystemExit) as e:
        test.stop(e)
        _listener.on_ignored_by_condition(test, e)
    except Exception as e:
        test.stop(e)
        _listener.on_broken(test, e)
    return False


def _check_data_providers(suite: TestSuite):
    """
    Checks if any test use data-provider, this provider is in collected list
    :param suite: TestSuite
    :return: None
    :raises UnknownProviderName if test uses name what is not in providers list
    """
    all_data_providers = [test.provider for group in suite.groups.values() for test in group.tests if test.provider]
    if not all_data_providers:
        return
    is_all_provider_known = all([provider in suite.providers for provider in all_data_providers])
    if not is_all_provider_known:
        name_ = [provider for provider in all_data_providers if provider not in suite.providers]
        raise UnknownProviderName(f'Cant find provider with name(s) {name_}. '
                                  f'You must have method with @data annotation in this package!')


def _run_before(test_case: TestCase):
    """
    Runs all fixtures before test case. If failed - set flaf is_before_failed to True
    :param test_case: Test or TestGroup
    :return: None
    """
    for before in test_case.before:
        result = _run_fixture(before, 'before', test_case.name)
        if result:
            test_case.is_before_failed = True


def _run_after(test_case: TestCase):
    """
    Runs all fixtures after test or group
    :param test_case: Test or TestGroup
    :return: None
    """
    if not test_case.always_run_after and test_case.is_before_failed:
        return
    for after in test_case.after:
        _run_fixture(after, 'after', test_case.name)


def _run_fixture(func: Callable, fixture_type: str, group_name: str) -> bool:
    """
    Runs fixture and returns result of the run
    :param func: fixture-function (marked @before etc.)
    :param fixture_type: type of the fixture to use in listener (before, after ...)
    :param group_name: name of the group
    :return: False if fixture not failed, True otherwise
    """
    is_failed: bool = False
    try:
        func()
    except Exception as error:
        _listener.on_fixture_failed(group_name, fixture_type, error)
        is_failed = True
    return is_failed



from time import time


class Timer:
    """
    Simple class for checking time
    """

    def __init__(self):
        self.start_time: float = -1
        self.end_time: float = -1
        self.duration: float = -1

    def start(self):
        self.start_time = time()

    def stop(self):
        self.end_time = time()
        self.duration = self.end_time - self.start_time

    def reset(self):
        self.start_time = -1
        self.end_time = -1
        self.duration = -1



from os import path
from sys import stderr
from sys import _getframe
from inspect import signature
from inspect import getsource
from inspect import isfunction
from typing import Callable, Any, Iterable, Tuple, Union, Sequence, Container

from .exceptions import *
from .classes.basic_test import Test
from .classes.data_file import DataFile
from .classes.basic_suite import TestSuite
from .helpers.others import is_file_exists, fake


def test(*args, enabled: bool = True, name: str = None, description: str = None, data_provider: str = None,
         retries: int = 1, groups: Tuple[str] = None, priority: int = 0, timeout: int = 0, only_if: Callable = None):
    """
    The annotation that marks a function in a module as a test, does not work with classes and class methods and with
    functions, that take an argument (except using of data provider).
    :param args: parameters, in which a function may come if the method is marked just with  @test.
    :param enabled: is the flag of the active test, if False then the test does not fall into the run and all its other
    settings are ignored.
    :param name: the name of the test, but if there is no name, then the name is the function name
    :param description: Test description, if None will be taken from function documentation. If there are description
    and documentation, then the parameter has the advantage, to wit this is in use, otherwise, documentation will be
    taken from documentation of the function (Test).
    :param data_provider: is the string name of data provider, which is not need to be in current module with test, the
    main is that it was found during assembling of test entities. If not found, the exception UnknownProviderName will
    be raised.
    :param retries: is the total amount of attempts of run test, this is the number of how many times the test will be
    run again in case of errors. If the test is successful, no more attempts are made, fixtures before and after the
    test are run just 1 time!
    :param groups: is the list of group names, to which the test will be assigned, if it empty, group automatically
    creates with the name of the module. This parameter allows to group tests from different modules to one run.
    :param priority: is the priority, for organization of tests execution order. The greater is 0, as parameter high the
    execution of the test will be done later.
    :param timeout is the number of seconds during waiting the test ends. If the test had not end, then
    TestBrokenException exception will raised, and the thread, in which the test is executing, will be interrupted.
    Due to a possible memory leak, it should be used only when there is a special need.
    :param only_if: accepts a function that will be called before the test starts and the test will only be launched if
    it returns True. Should be taken to filter tests, for instance in relation to the operating system used.
    :return: fake
    """
    if not enabled:
        return fake

    def real_decorator(func: Callable[[], None]):
        if not data_provider:
            __check_is_function_without_args(func, 'test')
        else:
            __check_is_function_for_provider(func)
        name_ = name if name else func.__name__
        _check_func_for_soft_assert(func)
        if only_if is not None and not callable(only_if):
            raise ValueError('Only_if parameter of @test annotation must be a function, returning True or False!')
        nonlocal groups
        if not groups:
            groups = [func.__module__]
        else:
            if type(groups) not in (list, tuple, set):
                raise ValueError('Group parameter of @test annotation must be a tuple of strings (Tuple[str])!')
        for group in groups:
            test_object = Test(name_, func)
            test_object.only_if = only_if
            test_object.retries = retries
            test_object.priority = priority
            if description:
                test_object.description = description
            if timeout:
                test_object.timeout = int(timeout)
                if test_object.timeout < 0:
                    test_object.timeout = 0
            if data_provider:
                test_object.provider = data_provider
            TestSuite.get_instance().get_or_create(group).add_test(test_object)
        return fake

    if args:
        return real_decorator(args[0])
    return real_decorator


def provider(*args, enabled: bool = True, name: str = None, cached: bool = False):
    """
    The annotation that marks a data provider, that is, a function that supplies data to a test. Such a function should
    return Iterable or Sequence, otherwise will be an error. It is not possible at compile time to determine if the
    function returns the correct type, so an exception with the wrong type will be thrown at runtime. Exception tests
    with such provider are added to ignored.
    :param args: are parameters in which a function may come if the method is marked simply with @data
    :param enabled: the flag of the active provider, if False, then it does not fall into the list of providers and all
    its other settings are ignored
    :param name: is the name, if not specified, then takes the name of the function. By this name, tests are searched by
    the provider, therefore only unique names are allowed. Duplicate name throws DuplicateNameException
    :param cached: flag to save all provider data in memory and use it on second use. Can be useful only when provider
    used more than once in test-suite and you do not want to get data again from some source like filesystem or db.
    WARNING! Cache use memory, so it can take a lot of it for big data volumes.
    :return: fake
    :raises: DuplicateNameException if provider with such name is already exists
    :raises: WrongAnnotationPlacement if @data annotation used on function without return or yield statements
    """
    if not enabled:
        return fake

    def real_decorator(func: Callable[[None], Iterable]):
        __check_is_function_without_args(func, 'data')
        if not _has_yield_or_return(func):
            raise WrongAnnotationPlacement(f'Function marked with @data must returns or yields Iterable!')
        name_ = name if name else func.__name__
        providers = TestSuite.get_instance().providers
        if name_ in providers:
            raise DuplicateNameException(f'Provider with name "{name_}" already exists! Only unique names allowed!')
        providers[name_] = func
        nonlocal cached
        if cached:
            TestSuite.get_instance().cached.append(name_)
        return fake

    if args:
        return real_decorator(args[0])
    return real_decorator


def before(*args, group_name: str = None):
    """
    It marks the function as mandatory to run before each module/group test.
    :param group_name: the name of the group before which test the function will be executed. If no group name is
    specified, a group is automatically created with the module name. A function does not have to be in the same module
    as the tests.
    :return: fake
    """

    def real_decorator(func: Callable[[], None]):
        __check_is_function_without_args(func, 'before')
        group = group_name if group_name else func.__module__
        TestSuite.get_instance().get_or_create(group).add_before_test(func)
        return fake

    if args:
        return real_decorator(args[0])
    return real_decorator


def after(*args, group_name: str = None):
    """
    It marks the function as mandatory to run after each module/group test. If there are functions running before the
    test (@before) and they failed, then these functions will not start!
    :param group_name: the name of the group after each test of which the function will be executed. If the group name
    is not specified, a group with the module name is automatically created. A function does not have to be in the same
    module as the tests.
    :return: fake
    """

    def real_decorator(func: Callable[[], None]):
        __check_is_function_without_args(func, 'after')
        group = group_name if group_name else func.__module__
        TestSuite.get_instance().get_or_create(group).add_after_test(func)
        return fake

    if args:
        return real_decorator(args[0])
    return real_decorator


def before_group(*args, name: str = None):
    """
    It marks the function as mandatory to run before the module/group is executed, that is, it is executed once before
    all module/group tests are run.
    :param name: is the name of the module or group, before which tests the function will be executed once. If no name
    is specified, then the name of the current module where the annotation is used is taken.
    :return: fake
    """

    def real_decorator(func: Callable[[], None]):
        __check_is_function_without_args(func, 'before_module')
        group = name if name else func.__module__
        TestSuite.get_instance().get_or_create(group).add_before(func)
        return fake

    if args:
        return real_decorator(args[0])
    return real_decorator


def after_group(*args, name: str = None, always_run: bool = False):
    """
    It marks the function as mandatory for run after the module/group is executed, that is, it is executed once after
    all module/group tests have been completed. If there is a function running before the module/group
    (@before_module) and it failed, then this function will not be launched if the always_run=True flag is not used.
    With this flag, the function ignores the results of preliminary functions and always starts.
    :param name: is the name of the module or group after which tests will be executed once the function. If no name is
    specified, then the name of the current module where the annotation is used is taken
    :param args: are parameters in which a function may come if the method is marked simply by @after_module
    :param always_run: is the function start flag, regardless of the result of the preliminary functions. If True, it
    will be launched anyway
    :return: fake
    """

    def real_decorator(func: Callable[[], None]):
        __check_is_function_without_args(func, 'after_module')
        group = name if name else func.__module__
        TestSuite.get_instance().get_or_create(group).add_after(func)
        if always_run:
            TestSuite.get_instance().get_or_create(group).always_run_after = True
        return fake

    if args:
        return real_decorator(args[0])
    return real_decorator


def before_suite(func: Callable[[], None]):
    """
    Marks the function as mandatory to run before executing the entire test-suite, that is, it is performed
    once at the very beginning of testing.
    :param func: is the function that does not take any argument
    :return: None
    """
    __check_is_function_without_args(func, 'before_suite')
    TestSuite.get_instance().add_before(func)


def after_suite(*args, always_run: bool = False):
    """
    It marks the function as mandatory for the run after the entire test run (test suite) has been completed, that is,
    it is performed once at the very end of the test after all groups and tests. If there are functions before the whole
    run (@before_suite) and they failed, then this function will not be executed, except when using the
    always_run=True flag. In this case, it will always be launched.
    :param args: are parameters in which a function may come if the method is marked simply @after_suite
    :param always_run: is the function start flag, regardless of the result of the preliminary functions. If True, it
    will be launched anyway
    :return: fake
    """

    def real_decorator(func: Callable[[], None]):
        __check_is_function_without_args(func, 'after_suite')
        TestSuite.get_instance().add_after(func)
        if always_run:
            TestSuite.always_run_after = True
        return fake

    if args:
        return real_decorator(args[0])
    return real_decorator


def _has_yield_or_return(function: Callable) -> bool:
    code = getsource(function)
    return ' return ' in code or ' yield ' in code


def _check_func_for_soft_assert(func):
    try:
        code = getsource(func)
        is_soft_assert_there = 'SoftAssert(' in code
        if not is_soft_assert_there:
            return
        if 'assert_all()' not in code:
            print(f'WARNING! Function {func.__module__}.{func.__name__} marked with @test seems to contains SoftAssert '
                  f'object without calling assert_all()!', file=stderr)
    except Exception:
        # Consciously ignore it, just check for a warning, this is not critical
        pass


def __check_is_function_without_args(func: Callable, annotation_name: str):
    """
    Checking that the annotation is above the function without arguments, it is not intended to use annotations with
    classes and / or with their methods.
    :param func: is the function to test
    :param annotation_name: is the name of annotation (for errors)
    :return: None
    :raises: WrongAnnotationPlacement
    """
    if not isfunction(func) or signature(func).parameters:
        raise WrongAnnotationPlacement(
            f"Annotation '{annotation_name}' must be used only with no-argument functions! Its not supposed to work "
            f"with classes or class methods!")


def __check_is_function_for_provider(func: Callable[[Any], None]):
    """
    Check that the function is suitable to accept values (use a data provider), that is, it has exactly 1 argument.
    :param func: is the function
    :return: None
    :raises: WrongAnnotationPlacement
    """
    if not isfunction(func) or not signature(func).parameters:
        raise WrongAnnotationPlacement(f"Function '{func.__name__}' marked with data_provider has no argument!")
    if len(signature(func).parameters) > 1:
        raise WrongAnnotationPlacement(f"Function '{func.__name__}' marked with data_provider "
                                       f"has more than 1 argument!")


def DATA_FILE(file_path: str, provider_name: str = None, cached: bool = False, encoding: str = 'UTF-8',
              map_function: Callable = None):
    """
    Function to use text file as data provider for test. Reads file lazily, do not get it to memory.
    The function name explicitly stays uppercase for user to pay attention to it.
    User must call it at the global module namespace, but not at fixtures or in tests!
    :param provider_name: name of the data-provider for use it in test, if not specified the file_path be used as name
    :param cached: flag to cache values for using it more than once
    :param file_path: file name or path-to-file with name, it can be full or relative path, but it must be "visible"
    (accessible from module, where it is declared)
    :param encoding: encoding of the text file (default UTF-8)
    :param map_function: function, which map line from text file
    :return: None
    :raises: ValueError if file is not exists!
    """

    def wrapper():
        return DataFile(real_path, encoding=encoding, map_function=map_function)

    if provider_name is None:
        provider_name = file_path
    try:
        # Get last frame to verify file-path
        frame = _getframe(1)
        assert frame  # It can't be no last frame!
        first_path = path.split(frame.f_globals['__file__'])[0]
        real_path = path.join(first_path, file_path)
        if not is_file_exists(real_path):
            raise ValueError(f'Cant find file! Is file "{real_path}" exists?')
        provider(name=provider_name, cached=cached)(wrapper)
    finally:
        del frame


def CONTAINER(value: Union[Sequence, Iterable, Container], name: str = None):
    """
    Sugar for simplify providing data, use it when provider is simple and can be written in one-liner, like list
    comprehension or generator expression.
    The function name explicitly stays uppercase for user to pay attention to it!
    User must call it at the global module namespace, but not at fixtures or in tests!
    :param value: sequence/iterable or any object you can use with for
    :param name: name for the provider, if empty then 'container' will be used as name
    :return: None
    """

    def _():
        return value

    name = name if name is not None else 'container'
    provider(name=name)(_)
